/* Known bugs (probably in StormC):
 (a) problems exiting concurrently running instances (a la Report+).
 (b) resource leakage (unknown cause) (a la Report+).

Possible improvements:
 support for BTCS character formats.
 condensation of empty inventory slots.
 'of's, 'to's should be vertically centred.
 enforcing that "current"s cannot be higher than "maximum"s.
 support doubleclick of listbrowser nodes.
 linting.
 to not allow "player"s to edit "monster/illusion" fields, and vice
  versa.
 information to include for items: nominal price, classes allowed to
  use, games it is allowed in (eg. BT1 only), armour improvement.
 use "image descriptions" (eg. dragon, etc.) for monster/illusion images,
  instead of numbers.
 allow editing of "items" file (BT1/2 only).
 allow editing of saved games (BT2/3 only).
 BT3: when changing class/sex, adjust image as appropriate.
 BT3: let user select which character to load/save.
 BT3: perhaps ghosted (ie. "current") attributes should mirror the
  editable (ie. "maximum") ones?
 cursor keys to scroll spells listbrowser up and down.
 automatically convert BT1 names, and BT2 player names, to uppercase.
 not to bother warning the user if saving as a new file, or overwriting
  a 1-character file.
 load and/or save Faery Tale Adventure characters.
 use Bard's Tale font.
 use Bard's Tale pointer.

Unknown file encodings:
 how are bank accounts stored in BT2?
 how are charged items (eg. Dragonshield) stored in BT1?

#include "lint.h" */

#include <exec/types.h>
#include <exec/nodes.h>
#include <exec/memory.h>
#include <exec/execbase.h>
#include <dos/dos.h>
#include <dos/dostags.h>
#include <dos/dosextens.h>
#include <graphics/gfx.h>
#include <graphics/gfxbase.h>
#include <graphics/displayinfo.h>
#include <workbench/icon.h>
#include <intuition/intuition.h>
#include <intuition/intuitionbase.h>
#include <libraries/asl.h>
#include <libraries/gadtools.h>
#define ALL_REACTION_CLASSES
#define ALL_REACTION_MACROS
#include <reaction/reaction.h>
#include <utility/tagitem.h>
#include <workbench/workbench.h>       /* struct DiskObject */

#include <proto/exec.h>
#include <proto/dos.h>
#include <proto/asl.h>
#include <proto/gadtools.h>
#include <proto/graphics.h>
#include <proto/icon.h>
#include <proto/intuition.h>
#include <proto/utility.h>
#include <clib/alib_protos.h>

#include <proto/chooser.h>
#include <proto/listbrowser.h>
#include <proto/resource.h>
#include <proto/space.h>
#include <proto/string.h>

#include <ctype.h>
#include <stdio.h>                     /* FILE, printf() */
#include <stdlib.h>                    /* EXIT_SUCCESS, EXIT_FAILURE */
#include <string.h>
#include <assert.h>

#include "gui.h"

/* 1. KEYWORDS -----------------------------------------------------------

Defined by AmigaOS are: IMPORT, UBYTE, UWORD, ULONG. */

#define AUTO                   auto   /* automatic variables */
#define MODULE                 static /* external static (file-scope) */
#define PERSIST                static /* internal static (function-scope) */
#define DISCARD                (void) /* discarded return values */
#define elif                   else if
#define acase                  break; case
#define adefault               break; default
#define EXPORT

#define PLATFORM_AMIGA   0
#define PLATFORM_IBM     1

#define PopUpEnd         ChooserEnd
#define TextEditorObject NewObject(TEXTEDITOR_GetClass(), NULL
#define TextEditorEnd    TAG_END)

#ifndef __LCLINT__
    #define ASM    __asm
    #define REG(x) register __ ## x
#endif

#ifdef __STORM__
    typedef char               FLAG;
    typedef char               SBYTE;   /*  8-bit signed quantity (replaces Amiga BYTE) */
    typedef short              SWORD;   /* 16-bit signed quantity (replaces Amiga WORD) */
    typedef long               SLONG;   /* 32-bit signed quantity (same as LONG) */
    typedef long               SHUGE;
    typedef unsigned long long UHUGE;
    #define EXIT_SUCCESS  0
    #define EXIT_FAILURE 20
#endif

// 2. DEFINES ------------------------------------------------------------

#define OS_ANY  0
#define OS_12  33
#define OS_13  34
#define OS_20  36
#define OS_204 37
#define OS_21  38
#define OS_30  39
#define OS_31  40
#define OS_32  41 // no such thing really of course
#define OS_35  44
#define OS_39  45
// what version is OS4.0?

#define SCAN_ESCAPE      69
#define SCAN_UP          76
#define SCAN_DOWN        77
#define SCAN_HELP        95

#define BT1                   0
#define BT2                   1
#define BT3                   2

#define CLASSGAD_ARCHMAGE     0
#define CLASSGAD_BARD         1
#define CLASSGAD_CHRONOMANCER 2
#define CLASSGAD_CONJURER     3
#define CLASSGAD_GEOMANCER    4
#define CLASSGAD_HUNTER       5
#define CLASSGAD_ILLUSION     6
#define CLASSGAD_MAGICIAN     7
#define CLASSGAD_MONK         8
#define CLASSGAD_MONSTER      9
#define CLASSGAD_PALADIN     10
#define CLASSGAD_ROGUE       11
#define CLASSGAD_SORCERER    12
#define CLASSGAD_WARRIOR     13
#define CLASSGAD_WIZARD      14

#define RACEGAD_HUMAN         6

#define MN_PROJECT        0
#define MN_SETTINGS       1
#define MN_HELP           2
#define IN_NEW            0
#define IN_OPEN           2
#define IN_SAVEAS         4
#define IN_QUIT           6
#define IN_APPEND         0
#define IN_ABOUT          0

#define SPELLS          124 // counting from zero

/* Module and global variables are initialized to zero by default.
   Any variables used in GetAttr() calls must be ULONGs. */
MODULE ULONG statusgad, race, theclass,
             curst, curiq, curdx, curcn, curlk,
             maxst, maxiq, maxdx, maxcn, maxlk,
             armour,
             curlev, maxlev,
             sorc, conj, magi, wiza, arch,
             songs, attacks, image, special, damage,
             cursor,
             game = BT1,
             equip[12], item[12], quantity[12],
             maxhp, curhp, maxsp, cursp,
             xp, gp,
             platform, sex,
             spell[SPELLS + 1];

MODULE struct NewMenu NewMenu[] =
{   { NM_TITLE, "Project",            0 , 0,                                                0, 0}, //  0
    {  NM_ITEM, "New",               "N", 0,                                                0, 0}, //  1
    {  NM_ITEM, NM_BARLABEL,          0 , 0,                                                0, 0}, //  2
    {  NM_ITEM, "Open...",           "O", 0,                                                0, 0}, //  3
    {  NM_ITEM, NM_BARLABEL,          0 , 0,                                                0, 0}, //  5
    {  NM_ITEM, "Save...",           "S", 0,                                                0, 0}, //  7
    {  NM_ITEM, NM_BARLABEL,          0 , 0,                                                0, 0}, //  8
    {  NM_ITEM, "Quit",              "Q", 0,                                                0, 0}, //  9
    { NM_TITLE, "Settings",           0 , 0,                                                0, 0}, // 10
    {  NM_ITEM, "Append EOF?",       "E", CHECKIT | MENUTOGGLE | CHECKED | NM_ITEMDISABLED, 0, 0}, // 11
    { NM_TITLE, "Help",               0 , 0,                                                0, 0}, // 12
    {  NM_ITEM, "About...",          "?", 0,                                                0, 0}, // 13
    {   NM_END, NULL,                 0 , 0,                                                0, 0}  // 14
};

MODULE struct EasyStruct EasyStruct =
{   sizeof(struct EasyStruct),
    0,
    "BTCE: Warning",
    "All other characters in %d will be destroyed!",
    "Confirm|Cancel"
};

IMPORT struct ExecBase*        SysBase; // optional under StormC

#ifdef __STORM__
    IMPORT struct Library*     VersionBase;
#endif

RESOURCEFILE                   ResourcePtr;
struct Gadget**                GadgetsPtr;

MODULE struct List             ItemsList,
                               SpellsList;
MODULE struct RDArgs*          ArgsPtr           /* = NULL */ ;
MODULE struct Menu*            MenuPtr           /* = NULL */ ;
MODULE struct Screen*          ScreenPtr         /* = NULL */ ;
MODULE struct Window*          MainWindowPtr     /* = NULL */ ;
MODULE struct VisualInfo*      VisualInfoPtr     /* = NULL */ ;
MODULE Object*                 WinObject         /* = NULL */ ; // note that WindowObject is a reserved macro
MODULE TEXT                    path[512 + 1]     = "",
                               pathname[768 + 1] = "", // can hold path + filename
                               filename[256 + 1] = "",
                               charname[14 + 1]  = "";
MODULE UBYTE                   IOBuffer[123];
MODULE FLAG                    append            = TRUE;

EXPORT struct Library         *ButtonBase,
                              *CheckBoxBase,
                              *ChooserBase,
                              *FuelGaugeBase,
                              *IntegerBase,
                              *LabelBase,
                              *LayoutBase,
                              *ListBrowserBase,
                              *ResourceBase,
                              *SpaceBase,
                              *StringBase,
                              *WindowBase;

MODULE ULONG gad_st[12] =
{   GID_ST1,
    GID_ST2,
    GID_ST3,
    GID_ST4,
    GID_ST5,
    GID_ST6,
    GID_ST7,
    GID_ST8,
    GID_ST9,
    GID_ST10,
    GID_ST11,
    GID_ST12
}, gad_in[12] =
{   GID_IN29,
    GID_IN30,
    GID_IN31,
    GID_IN32,
    GID_IN33,
    GID_IN34,
    GID_IN35,
    GID_IN36,
    GID_IN37,
    GID_IN38,
    GID_IN39,
    GID_IN40
}, gad_cb[12] =
{   GID_CB1,
    GID_CB2,
    GID_CB3,
    GID_CB4,
    GID_CB5,
    GID_CB6,
    GID_CB7,
    GID_CB8,
    GID_CB9,
    GID_CB10,
    GID_CB11,
    GID_CB12
}, gad_bu[12] =
{   GID_BU2,
    GID_BU3,
    GID_BU4,
    GID_BU5,
    GID_BU6,
    GID_BU7,
    GID_BU8,
    GID_BU9,
    GID_BU10,
    GID_BU11,
    GID_BU12,
    GID_BU13
};

#define ITEMS 291 // actual number of items available (not counting "Nothing")

MODULE struct
{   UBYTE  bt1, bt2, bt3;
    STRPTR name;
} items[ITEMS + 1] =
{ 0,   0,   0, "Nothing",
  0,   0, 117, "Acorn",
 62,  57,  57, "Admt Chain",
 51,   0,   0, "Admt Dagger",
 53,  51,  51, "Admt Gloves",
 52,  50,  50, "Admt Helm",
 54,   0,   0, "Admt Mace",
 64,  59,  59, "Admt Plate",
 63,   0,   0, "Admt Scale",
 50,  49,  49, "Admt Shield",
 49,  48,  48, "Admt Sword",
  0,  72,  72, "Ag's Arrows",
 58,  54,  54, "Ali's Carpet",
  0,   0, 210, "Angel's Harp",
  0,   0, 185, "Angel's Ring",
  0, 102, 102, "Angra's Eye",
  0, 101, 101, "Aram's Knife",
105,   0,   0, "Arc's Eye",
 85,   0,   0, "Arc's Hammer",
 66,   0,   0, "Arcshield",
  0,   0, 122, "Arefolia",
  0,  22,  22, "Arrows",
  0,   0, 124, "Arrows of Life",
  0,  74,  74, "Bard Bow",
 29,  29,  29, "Bardsword",
  0,   0, 136, "Belt of Alliria",
  0,   0, 159, "Black Arrows",
  0,   0, 114, "Black Lens",
 44,   0,   0, "Blood Axe",
  0,   0, 201, "Blood Mesh Robe",
  0,   0, 146, "Blue Rose",
  0,   0, 167, "Bolt",
  0,  53,  53, "Boomerang",
 65,  60,  60, "Bracers [4]",
 28,  28,  28, "Bracers [6]",
  0,  82,  82, "Breathring",
  3,   3,   3, "Broadsword",
 55,   0,   0, "Broom",
  0, 105, 105, "Brothers Fgn",
 10,  10,  10, "Buckler",
  0,  57,  57, "Bulldozer Fgn",
  0,   0, 125, "Canteen",
 13,  13,  13, "Chain Mail",
  0,   0, 222, "Cli Lyre",
  0,  30,  30, "Cold Horn",
 84,  78,  78, "Conjurstaff",
  0,   0, 135, "Crown of Truth",
  0,   0, 220, "Crystal Gem",
  0,   0, 137, "Crystal Key",
  0,   0, 112, "Crystal Lens",
113,  89,  89, "Crystal Sword",
  5,   5,   5, "Dagger",
  0, 118,   0, "Dagger!",
104,   0,   0, "Dag Stone",
 45,  44,  44, "Dayblade",
116,   0,   0, "Death Dagger",
  0,   0, 190, "Death Fgn",
  0,   0, 200, "Death Hammer",
  0,   0, 186, "Deathhorn",
101,  88,  88, "Deathring",
  0,  58,  58, "Death Stars",
  0,   0, 142, "Divine Halbard",
  0,   0, 189, "Dmnd Bracers",
 80,   0,   0, "Dmnd Dagger",
  0,   0, 214, "Dmnd Flail",
 81,  75,  75, "Dmnd Helm",
 90,  84,  84, "Dmnd Plate",
  0,   0, 181, "Dmnd Scale",
 79,  73,  73, "Dmnd Shield",
  0,   0, 219, "Dmnd Staff",
  0,   0, 213, "Dmnd Suit",
 78,   0,   0, "Dmnd Sword",
 41,   0,   0, "Dork Ring",
108,  98,  98, "Dragon Fgn",
 89,  83,  83, "Dragonshield",
 93,  87,  87, "Dragonwand",
  0,  93,  93, "Drums of Death",
  0, 106, 106, "Dynamite",
  0,   0, 217, "Eelskin Tunic",
  0,  76,  76, "Elf Boots",
 47,  46,  46, "Elf Cloak",
  0,   0, 183, "Eternal Torch",
 57,   0,   0, "Exorwand",
119,   0,   0, "Eye",
  0,   0, 204, "Familiar Fgn",
  0,   0, 153, "Ferofist's Helm",
 42,  41,  41, "Fin's Flute",
  0,   0, 131, "Firebrand",
 30,   0,   0, "Fire Horn",
  0,   0, 129, "Fire Spear",
 96,  91,  91, "Flame Horn",
  0,   0, 175, "Flame Knife",
  0,   0, 243, "Flare Crystal",
 22,   0,   0, "Flute",
 77,  71,  71, "Frost Horn",
 76,  70,  70, "Galt's Flute",
  0,   0, 194, "Galvanic Oboe",
 19,  19,  19, "Gauntlets",
 61,  27,  27, "Giant Fgn",
  0,   0, 245, "Gods' Blade",
 82,   0,   0, "Golem Fgn",
  0, 109, 109, "Grenade",
  7,   7,   7, "Halbard",
  0,   0, 152, "Hammer of Wrath",
 21,   0,   0, "Harp",
  0,   0, 195, "Harmonic Gem",
 48,  47,  47, "Hawkblade",
 75,   0,   0, "Heal Harp",
  0,   0, 177, "Heartseeker",
 17,  17,  17, "Helm",
  0,   0, 156, "Helm of Justice",
  0, 103, 103, "Herb Fgn",
  0,   0, 172, "Holy Avenger",
  0,   0, 244, "Holy Missile",
  0,   0, 132, "Holy Sword",
  0,   0, 182, "Holy TNT",
  0,   0, 248, "Horn of Gods",
  0,   0, 205, "Hourglass",
  0,   0, 246, "Hunter Blade",
  0,   0, 199, "Hunters Cloak",
  0,   0, 144, "I-ching",
  0,   0, 143, "Incense",
  0, 115,   0, "Item of Kazdek",
 43,  42,  42, "Kael's Axe",
  0,   0, 174, "Kali's Garrotte",
  0,  68,  68, "Kato's Bracer",
 94,   0,   0, "Kiel's Compass",
 39,   0,   0, "Lak's Lyre",
  2,   2,   2, "Lamp",
 12,  12,  12, "Leather Armor",
 18,  18,  18, "Leather Glvs.",
  0,   0, 164, "Left Key",
  0,   0, 165, "Lever",
118,   0,   0, "Lich Fgn",
  0,   0, 134, "Lightstar",
 31,  31,  31, "Lightwand",
  0,   8,   8, "Long Bow",
 92,   0,   0, "Lorehelm",
 60,  55,  55, "Luckshield",
  8,   0,   0, "Mace",
109,   0,   0, "Mage Fgn",
 68,  63,  63, "Mage Staff",
  0,   0, 203, "Mages Cloak",
  0,   0, 242, "Mages Glove",
 59,   0,   0, "Magic Mouth",
  0,   0, 150, "Magic Triangle",
 20,  20,  20, "Mandolin",
  0,  99,  99, "Mastermage Fgn",
120, 110, 110, "Master Key",
  0, 104, 104, "Master Wand",
  0,   0, 198, "Minstrels Glove",
  0,   0, 171, "Misericorde",
  0,  38,  38, "Molten Fgn",
117,   0,   0, "Mongo Fgn",
 88,   0,   0, "Mournblade",
  0,  43,  43, "Mthr Arrows",
 35,  35,  35, "Mthr Axe",
 25,  25,  25, "Mthr Chain",
 32,  32,  32, "Mthr Dagger",
 34,  34,  34, "Mthr Gloves",
 33,  33,  33, "Mthr Helm",
 36,   0,   0, "Mthr Mace",
 37,  37,  37, "Mthr Plate",
 26,  26,  26, "Mthr Scale",
 24,  24,  24, "Mthr Shield",
  0,   0, 240, "Mthr Suit",
 23,  23,  23, "Mthr Sword",
  0,   0, 119, "Nightspear",
  0, 111, 111, "Nospin Ring",
  0,   0, 166, "Nut",
 38,   0,   0, "Ogre Fgn",
106,  67,  67, "Ogrewand",
126,   0,   0, "Old Man Fgn",
112,   0,   0, "Onyx Key",
  0,   0, 184, "Oscon's Staff",
 99,  94,  94, "Pipes of Pan",
 15,  15,  15, "Plate Armor",
  0,   0, 192, "Poison Dagger",
 87,  81,  81, "Powerstaff",
 56,  52,  52, "Pureblade",
 67,  62,  62, "Pure Shield",
  0,   0, 215, "Purple Heart",
  0,   0, 149, "Rainbow Rose",
  0,   0, 176, "Red's Stiletto",
  0,   0, 163, "Right Key",
100,  95,  95, "Ring of Power",
  0,  80,  80, "Ring of Return",
  0,   0, 147, "Red Rose",
 16,  16,  16, "Robes",
 27,   0,   0, "Samurai Fgn",
 14,  14,  14, "Scale Armor",
  0,   0, 157, "Sceadu's Cloak",
  0, 119,   0, "Seg #1",
  0, 120,   0, "Seg #2",
  0, 121,   0, "Seg #3",
  0, 122,   0, "Seg #4",
  0, 123,   0, "Seg #5",
  0, 124,   0, "Seg #6",
  0, 125,   0, "Seg #7",
  0,   0, 158, "Shadelance",
  0,   0, 170, "Shadow Door",
  0,   0, 169, "Shadow Lock",
  0,   0, 173, "Shadowshiv",
  0,   0, 162, "Sheetmusic",
 40,  40,  40, "Shield Ring",
 46,  45,  45, "Shield Staff",
  4,   4,   4, "Short Sword",
  0,   0, 209, "Shrill Flute",
  0,  36,  36, "Shuriken",
123,   0,   0, "Silver Circle",
122,   0,   0, "Silver Square",
124,   0,   0, "Silver Triangle",
  0,  61,  61, "Slayer Fgn",
  0,   0, 113, "Smokey Lens",
  0,  96,  96, "Song Axe",
  0,   0, 202, "Soothing Balm",
  0,   0, 218, "Sorcerer's Hood",
 73,  69,  69, "Sorcerstaff",
 71,  66,  66, "Soul Mace",
  0,   0, 168, "Spanner",
  0,   0, 193, "Spark Blade",
103,   0,   0, "Spectre Mace",
127, 127,   0, "Spectre Snare",
  0,  21,  21, "Spear",
 95,  90,  90, "Speedboots",
  0,  39,  39, "Spell Spear",
  0,   0, 115, "Sphere of Lanatir",
 98,   0,   0, "Spiritdrum",
107,   0,   0, "Spirithelm",
  9,   9,   9, "Staff",
  0,   0, 247, "Staff of Gods",
 86,  79,  79, "Staff of Lor",
  0,   0, 187, "Staff of Mangar",
  0,   0, 141, "Steady Eye",
  0,   0, 139, "Stealth Arrows",
114, 108, 108, "Stoneblade",
  0,   0, 161, "Strifespear",
  0,   0, 207, "Surehand Amulet",
 74,   0,   0, "Sword of Pak",
  0, 113,   0, "Sword of Zar",
  0,   0, 138, "Tao Ring",
  0,   0, 188, "Tesla Ring",
  0,   0, 211, "The Book",
  0, 116,   0, "The Ring",
  0, 126,   0, "The Scepter",
 70,  65,  65, "Thief Dagger",
  0,   0, 208, "Thieves Dart",
  0,   0, 206, "Thieves Hood",
125,   0,   0, "Thor Fgn",
  0, 107, 107, "Thor's Hammer",
  0,   0, 191, "Thunder Sword",
  0,   0, 216, "Titan Bracers",
 83,   0,   0, "Titan Fgn",
  0,   0, 128, "Titan Helm",
  0,   0, 126, "Titan Plate",
  0,   0, 127, "Titan Shield",
  0,   0, 241, "Titan Suit",
  1,   1,   1, "Torch",
  0, 112,   0, "Torch!",
 11,  11,  11, "Tower Shield",
115,   0,   0, "Travelhelm",
  0,  97,  97, "Trick Brick",
110, 100, 100, "Troll Ring",
111,   0,   0, "Troll Staff",
  0,   0, 212, "Troth Lance",
  0, 117,   0, "Troy P.",
 97,   0,   0, "Truthdrum",
  0,   0, 120, "Tslotha's Head",
  0,   0, 121, "Tslotha's Heart",
  0,   0, 197, "Tung Plate",
  0,   0, 196, "Tung Shield",
  0,   0, 123, "Valarian's Bow",
  0,  77,  77, "Van Fgn",
  0, 114,   0, "Vial",
  0,   0, 221, "Wand of Force",
  0,   0, 133, "Wand of Fury",
  0,   0, 116, "Wand of Power",
  6,   6,   6, "War Axe",
 91,  85,  85, "Wargloves",
 69,  64,  64, "War Staff",
  0,   0, 160, "Werra's Shield",
  0,   0, 145, "White Rose",
  0,   0, 118, "Wineskin",
  0,   0, 130, "Willow Flute",
 72,   0,   0, "Wither Staff",
  0,  86,  86, "Wizhelm",
112,   0,   0, "Wizwand",
102,   0,   0, "Ybarrashield",
  0,   0, 148, "Yellow Rose",
  0,   0, 140, "Yellow Staff",
  0,   0, 223, "Youth Potion",
  0,  92,  92, "Zen Arrows"
};

MODULE struct
{   STRPTR longname;
    UBYTE  level,
           theclass,
           thebyte,
           thebit;
} spells[SPELLS + 1] =
{
"Animate Dead",                    4, CLASSGAD_WIZARD,       92, 128,
"Anti-Magic",                      5, CLASSGAD_MAGICIAN,     87,   2,
"Apport Arcane",                   7, CLASSGAD_CONJURER,     86, 128,
"Arboria -> Realm",                1, CLASSGAD_CHRONOMANCER, 94,  64,
"Area Enchant",                    2, CLASSGAD_MAGICIAN,     87, 128,
"Arc Fire",                        1, CLASSGAD_CONJURER,     84,  64,
"Batch Spell",                     2, CLASSGAD_ARCHMAGE,     93,  64,
"Baylor's Spell Bind",             5, CLASSGAD_WIZARD,       92,  64,
"Beyond Death",                    6, CLASSGAD_WIZARD,       92,   8,
"Brothers Kringle, The",           6, CLASSGAD_ARCHMAGE,     93,   4,
"Camaraderie",                     3, CLASSGAD_ARCHMAGE,     93,  32,
"Cat Eyes",                        4, CLASSGAD_SORCERER,     89,   1,
"Death Strike",                    7, CLASSGAD_MAGICIAN,     88,  16,
"Demon Bane",                      2, CLASSGAD_WIZARD,       91,   8,
"Demon Strike",                    7, CLASSGAD_WIZARD,       92,   2,
"Disbelieve",                      2, CLASSGAD_SORCERER,     89,  64,
"Dispossess",                      3, CLASSGAD_WIZARD,       91,   2,
"Disrupt Illusion",                5, CLASSGAD_SORCERER,     90,  64,
"Divine Intervention",             0, CLASSGAD_WARRIOR,      99,  16,
"Dragon Breath",                   4, CLASSGAD_MAGICIAN,     87,   4,
"Earth Dagger",                    1, CLASSGAD_GEOMANCER,    97,  32,
"Earth Elemental",                 2, CLASSGAD_GEOMANCER,    97,   2,
"Earth Maw",                       7, CLASSGAD_GEOMANCER,    99,  64,
"Earth Song",                      1, CLASSGAD_GEOMANCER,    97,  16,
"Earth Ward",                      1, CLASSGAD_GEOMANCER,    97,   8,
"Elik's Instant Slayer",           7, CLASSGAD_CONJURER,     86,  32,
"Elik's Instant Wolf",             4, CLASSGAD_CONJURER,     85,  64,
"Fanskar's Force Focus",           1, CLASSGAD_WIZARD,       91,  32,
"Fanskar's Night Lance",           4, CLASSGAD_ARCHMAGE,     93,  16,
"Far Death",                       4, CLASSGAD_CHRONOMANCER, 95,  16,
"Far Foe",                         7, CLASSGAD_CONJURER,     86,  64,
"Fatal Fist",                      7, CLASSGAD_CHRONOMANCER, 96,   1,
"Flame Column",                    3, CLASSGAD_WIZARD,       91,   4,
"Flesh Anew",                      6, CLASSGAD_CONJURER,     85,   4,
"Flesh Restore",                   4, CLASSGAD_CONJURER,     85,  32,
"Force of Tarjan",                 6, CLASSGAD_CHRONOMANCER, 96,  16,
"Freeze Foes",                     2, CLASSGAD_CONJURER,     84,  16,
"Frost Force",                     2, CLASSGAD_CHRONOMANCER, 94,  16,
"Gelidia -> Realm",                2, CLASSGAD_CHRONOMANCER, 94,   4,
"Gilles' Gills",                   0, CLASSGAD_WARRIOR,      99,  32,
"Giant Strength",                  5, CLASSGAD_MAGICIAN,     87,   1,
"Glacier Strike",                  5, CLASSGAD_GEOMANCER,    98,   4,
"God Fire",                        3, CLASSGAD_CHRONOMANCER, 94,   2,
"Gotterdamurung",                  0, CLASSGAD_WARRIOR,      99,   8,
"Grave Robber",                    6, CLASSGAD_CHRONOMANCER, 96,  32,
"Greater Revelation",              5, CLASSGAD_CONJURER,     85,  16,
"Greater Summoning",               6, CLASSGAD_WIZARD,       92,  16,
"Heal All",                        5, CLASSGAD_ARCHMAGE,     93,   8,
"Holy Water",                      2, CLASSGAD_MAGICIAN,     86,   2,
"Ice Storm",                       7, CLASSGAD_MAGICIAN,     88,   8,
"Identify",                        5, CLASSGAD_CHRONOMANCER, 95,   2,
"Kiel's Magic Compass",            2, CLASSGAD_CONJURER,     84,   8,
"Kinestia -> Realm",               4, CLASSGAD_CHRONOMANCER, 95,   4,
"Kylearan's Invisibility Spell",   3, CLASSGAD_SORCERER,     89,   4,
"Jolt Bolt",                       7, CLASSGAD_GEOMANCER,    99, 128,
"Levitation",                      3, CLASSGAD_CONJURER,     84,   1,
"Lesser Revelation",               3, CLASSGAD_CONJURER,     84,   2,
"Locate Traps",                    1, CLASSGAD_SORCERER,     89, 128,
"Lucencia -> Realm",               3, CLASSGAD_CHRONOMANCER, 95,  64,
"Luck Chant",                      4, CLASSGAD_CHRONOMANCER, 95,  32,
"Mage Flame",                      1, CLASSGAD_CONJURER,     84, 128,
"Mage Gauntlets",                  2, CLASSGAD_MAGICIAN,     86,   1,
"Mage Maelstrom",                  7, CLASSGAD_SORCERER,     90,   1,
"Magma Blast",                     6, CLASSGAD_GEOMANCER,    98,   1,
"Malefia -> Realm",                7, CLASSGAD_CHRONOMANCER, 97,  64,
"Major Levitation",                6, CLASSGAD_CONJURER,     85,   2,
"Mangar's Mallet",                 7, CLASSGAD_ARCHMAGE,     93,   2,
"Mangar's Mind Blade",             5, CLASSGAD_SORCERER,     90,  32,
"Mangar's Mind Jab",               1, CLASSGAD_SORCERER,     88,   2,
"Melee Men",                       1, CLASSGAD_ARCHMAGE,     93, 128,
"Oscon's Haltfoe",                 1, CLASSGAD_ARCHMAGE,     92,   1,
"Oscon's Ogre Strength",           3, CLASSGAD_MAGICIAN,     87,  32,
"Pathfinder",                      6, CLASSGAD_GEOMANCER,    98,   2,
"Petrify",                         3, CLASSGAD_GEOMANCER,    98, 128,
"Phase Blur",                      1, CLASSGAD_SORCERER,     88,   1,
"Phase Door",                      6, CLASSGAD_MAGICIAN,     88, 128,
"Preclusion",                      7, CLASSGAD_SORCERER,     91, 128,
"Prime Summoning",                 2, CLASSGAD_WIZARD,       91,  16,
"Quick Fix",                       1, CLASSGAD_MAGICIAN,     86,   8,
"Realm -> Arboria",                1, CLASSGAD_CHRONOMANCER, 94, 128,
"Realm -> Gelidia",                2, CLASSGAD_CHRONOMANCER, 94,   8,
"Realm -> Kinestia",               4, CLASSGAD_CHRONOMANCER, 95,   8,
"Realm -> Lucencia",               3, CLASSGAD_CHRONOMANCER, 95, 128,
"Realm -> Malefia",                7, CLASSGAD_CHRONOMANCER, 97, 128,
"Realm -> Tarmitia",               6, CLASSGAD_CHRONOMANCER, 96,   8,
"Realm -> Tenebrosia",             5, CLASSGAD_CHRONOMANCER, 96, 128,
"Regeneration",                    7, CLASSGAD_CONJURER,     85,   1,
"Restoration",                     7, CLASSGAD_MAGICIAN,     88,  32,
"Rimefang",                        7, CLASSGAD_SORCERER,     90,   4,
"Roscoe's Alert",                  3, CLASSGAD_GEOMANCER,    98,  64,
"Sanctuary",                       5, CLASSGAD_GEOMANCER,    98,   8,
"Sandstorm",                       4, CLASSGAD_GEOMANCER,    98,  16,
"Scry Site",                       1, CLASSGAD_MAGICIAN,     86,   4,
"Second Sight",                    3, CLASSGAD_SORCERER,     89,   2,
"Shadow Shield",                   7, CLASSGAD_CHRONOMANCER, 96,   2,
"Shock Sphere",                    5, CLASSGAD_CONJURER,     85,   8,
"Sorcerer Sight",                  6, CLASSGAD_SORCERER,     90,   8,
"Spectre Touch",                   4, CLASSGAD_MAGICIAN,     87,   8,
"Star Flare",                      3, CLASSGAD_MAGICIAN,     87,  16,
"Stone to Flesh",                  7, CLASSGAD_MAGICIAN,     88,   4,
"Storal's Soul Whip",              5, CLASSGAD_WIZARD,       92,  32,
"Stun",                            3, CLASSGAD_CHRONOMANCER, 94,   1,
"Succor Song",                     4, CLASSGAD_GEOMANCER,    98,  32,
"Summon Elemental",                1, CLASSGAD_WIZARD,       91,  64,
"Summon Herb",                     4, CLASSGAD_WIZARD,       91,   1,
"Tarmitia -> Realm",               6, CLASSGAD_CHRONOMANCER, 96,   4,
"Tenebrosia -> Realm",             5, CLASSGAD_CHRONOMANCER, 96,  64,
"Trebuchet",                       2, CLASSGAD_GEOMANCER,    97,   4,
"Trap Zap",                        1, CLASSGAD_CONJURER,     84,  32,
"Vitality",                        1, CLASSGAD_CHRONOMANCER, 93,   1,
"Vorpal Plating",                  1, CLASSGAD_MAGICIAN,     86,  16,
"Wacum's Wizard War",              7, CLASSGAD_WIZARD,       92,   4,
"Wall Warp",                       2, CLASSGAD_GEOMANCER,    97,   1,
"Warstrike",                       3, CLASSGAD_CONJURER,     85, 128,
"Wind Dragon",                     4, CLASSGAD_SORCERER,     90, 128,
"Wind Giant",                      6, CLASSGAD_SORCERER,     90,  16,
"Wind Hero",                       7, CLASSGAD_SORCERER,     90,   2,
"Wind Ogre",                       3, CLASSGAD_SORCERER,     89,   8,
"Wind Warrior",                    2, CLASSGAD_SORCERER,     89,  32,
"Witherfist",                      2, CLASSGAD_CHRONOMANCER, 94,  32,
"Word of Fear",                    2, CLASSGAD_SORCERER,     89,  16,
"Word of Healing",                 2, CLASSGAD_CONJURER,     84,   4,
"Ybarra's Mystical Coat of Armor", 6, CLASSGAD_MAGICIAN,     88,  64,
"Ybarra's Mystic Shield",          3, CLASSGAD_MAGICIAN,     87,  64,
"Youth",                           5, CLASSGAD_CHRONOMANCER, 95,   1
};

MODULE FLAG load(void);
MODULE void load_bt1(void);
MODULE void load_bt2(void);
MODULE void load_bt3(void);
MODULE LONG getsize(void);
MODULE SLONG getlong(void);
MODULE SLONG getilong(void);
MODULE SWORD getword(void);
MODULE SWORD getiword(void);
MODULE UBYTE getfatbyte(void);
MODULE UBYTE getbyte(void);
MODULE void bt1_to_bt2(void);
MODULE void bt1_to_bt3(void);
MODULE void bt2_to_bt1(void);
MODULE void bt2_to_bt3(void);
MODULE void bt3_to_bt1(void);
MODULE void bt3_to_bt2(void);
MODULE void refreshgadgets(ULONG mode);
MODULE void lockscreen(void);
MODULE void unlockscreen(void);
MODULE void cleanexit(SBYTE rc);
MODULE void clearkybd(void);
MODULE void handlemenus(UWORD code);
MODULE void rq(STRPTR message);
MODULE void writeout(SLONG length);
MODULE void open(void);
MODULE void save(void);
MODULE void readgadgets(void);
MODULE FLAG asl(FLAG save);
MODULE void about(void);
MODULE void itemwindow(ULONG whichitem);
MODULE void loop(ULONG gid);
MODULE void filetovar_race(UBYTE thebyte);
MODULE void filetovar_class(UBYTE thebyte);
MODULE void filetovar_status(UBYTE thebyte);
MODULE UBYTE vartofile_race(void);
MODULE UBYTE vartofile_class(void);
MODULE UBYTE vartofile_status(void);
MODULE void item_loop(ULONG gid, UWORD code, ULONG whichitem);
/* MODULE void condense(void); */
MODULE void project_new(void);
MODULE void save_bt1(void);
MODULE void save_bt2(void);
MODULE void save_bt3(void);
MODULE void make_charname(void);
MODULE LONG warn(void);

MODULE void readgadgets(void)
{   ULONG        i;
    STRPTR       stringptr;
    struct Node* NodePtr;
    struct List* BrowserListPtr;

    // general
    DISCARD GetAttr(STRINGA_TextVal,    GadgetsPtr[GID_ST13], (ULONG *) &stringptr);
    strcpy(charname, stringptr);
    DISCARD GetAttr(CHOOSER_Selected,   GadgetsPtr[GID_CH1],  (ULONG *) &game);
    DISCARD GetAttr(CHOOSER_Selected,   GadgetsPtr[GID_CH6],  (ULONG *) &platform);
    DISCARD GetAttr(LISTBROWSER_Selected, GadgetsPtr[GID_LB3], (ULONG *) &theclass);
    DISCARD GetAttr(INTEGER_Number,     GadgetsPtr[GID_IN1],  (ULONG *) &gp);
    DISCARD GetAttr(INTEGER_Number,     GadgetsPtr[GID_IN2],  (ULONG *) &curhp);
    DISCARD GetAttr(INTEGER_Number,     GadgetsPtr[GID_IN3],  (ULONG *) &maxhp);
    DISCARD GetAttr(INTEGER_Number,     GadgetsPtr[GID_IN4],  (ULONG *) &cursp);
    DISCARD GetAttr(INTEGER_Number,     GadgetsPtr[GID_IN5],  (ULONG *) &maxsp);
    DISCARD GetAttr(CHOOSER_Selected,   GadgetsPtr[GID_CH5],  (ULONG *) &statusgad);

    // attributes
    DISCARD GetAttr(INTEGER_Number,     GadgetsPtr[GID_IN6],  (ULONG *) &curst);
    DISCARD GetAttr(INTEGER_Number,     GadgetsPtr[GID_IN7],  (ULONG *) &curiq);
    DISCARD GetAttr(INTEGER_Number,     GadgetsPtr[GID_IN8],  (ULONG *) &curdx);
    DISCARD GetAttr(INTEGER_Number,     GadgetsPtr[GID_IN9],  (ULONG *) &curcn);
    DISCARD GetAttr(INTEGER_Number,     GadgetsPtr[GID_IN10], (ULONG *) &curlk);
    DISCARD GetAttr(INTEGER_Number,     GadgetsPtr[GID_IN11], (ULONG *) &maxst);
    DISCARD GetAttr(INTEGER_Number,     GadgetsPtr[GID_IN12], (ULONG *) &maxiq);
    DISCARD GetAttr(INTEGER_Number,     GadgetsPtr[GID_IN13], (ULONG *) &maxdx);
    DISCARD GetAttr(INTEGER_Number,     GadgetsPtr[GID_IN14], (ULONG *) &maxcn);
    DISCARD GetAttr(INTEGER_Number,     GadgetsPtr[GID_IN15], (ULONG *) &maxlk);
    DISCARD GetAttr(INTEGER_Number,     GadgetsPtr[GID_IN16], (ULONG *) &magi);
    DISCARD GetAttr(INTEGER_Number,     GadgetsPtr[GID_IN17], (ULONG *) &conj);
    DISCARD GetAttr(INTEGER_Number,     GadgetsPtr[GID_IN18], (ULONG *) &sorc);
    DISCARD GetAttr(INTEGER_Number,     GadgetsPtr[GID_IN19], (ULONG *) &wiza);
    DISCARD GetAttr(INTEGER_Number,     GadgetsPtr[GID_IN20], (ULONG *) &arch);

    // illusions/monsters
    DISCARD GetAttr(INTEGER_Number,     GadgetsPtr[GID_IN21], (ULONG *) &armour);
    DISCARD GetAttr(INTEGER_Number,     GadgetsPtr[GID_IN22], (ULONG *) &attacks);
    DISCARD GetAttr(INTEGER_Number,     GadgetsPtr[GID_IN23], (ULONG *) &damage);
    DISCARD GetAttr(INTEGER_Number,     GadgetsPtr[GID_IN24], (ULONG *) &image);
    DISCARD GetAttr(CHOOSER_Selected,   GadgetsPtr[GID_CH3],  (ULONG *) &special);

    // players
    DISCARD GetAttr(CHOOSER_Selected,   GadgetsPtr[GID_CH4],   (ULONG *) &race);
    DISCARD GetAttr(INTEGER_Number,     GadgetsPtr[GID_IN25],  (ULONG *) &curlev);
    DISCARD GetAttr(INTEGER_Number,     GadgetsPtr[GID_IN26],  (ULONG *) &maxlev);
    DISCARD GetAttr(INTEGER_Number,     GadgetsPtr[GID_IN27],  (ULONG *) &songs);
    DISCARD GetAttr(INTEGER_Number,     GadgetsPtr[GID_IN28],  (ULONG *) &xp);
    DISCARD GetAttr(CHOOSER_Selected,   GadgetsPtr[GID_CH2],   (ULONG *) &sex);

    // class
    DISCARD GetAttr(LISTBROWSER_Selected, GadgetsPtr[GID_LB3], (ULONG *) &theclass);

    // items
    for (i = 0; i <= 7; i++)
    {   DISCARD GetAttr(GA_Selected,    GadgetsPtr[gad_cb[i]], (ULONG *) &equip[i]);
        DISCARD GetAttr(INTEGER_Number, GadgetsPtr[gad_in[i]], (ULONG *) &quantity[i]);
        // no need to read the item choosers
    }

/* why was this here!?
    // spells
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_LB2], MainWindowPtr, NULL,
        GA_Disabled, game != BT3,
    TAG_END); */

    DISCARD GetAttr(LISTBROWSER_Labels, GadgetsPtr[GID_LB2],   (ULONG *) &BrowserListPtr);
    assert(BrowserListPtr->lh_Head->ln_Succ); // the list is non-empty

    i = 0;
    // Walk the list
    for
    (   NodePtr = BrowserListPtr->lh_Head;
        NodePtr->ln_Succ;
        NodePtr = NodePtr->ln_Succ
    )
    {   DISCARD GetListBrowserNodeAttrs(NodePtr, LBNA_Checked, (ULONG *) &spell[i]);
        i++;
}   }

MODULE FLAG load(void)
{   LONG size;
    BPTR FileHandle /* = NULL */ ;

    size = getsize();
    if (size == 86)
    {   game = BT2;
        platform = PLATFORM_AMIGA;
    } elif (size == 96)
    {   game = BT1;
        platform = PLATFORM_AMIGA;
    } elif (size == 103)
    {   game = BT2;
        platform = PLATFORM_IBM;
    } elif (size == 109)
    {   game = BT1;
        platform = PLATFORM_IBM;
    } elif (size)
    {   if (size % 122 <= 1)
        {   game = BT3;
            platform = PLATFORM_AMIGA;
            size = 122;
        } elif (size % 120 <= 1)
        {   game = BT3;
            platform = PLATFORM_IBM;
            size = 120;
        } else
        {   DisplayBeep(NULL);
            return(FALSE);
    }   }
    else
    {   DisplayBeep(NULL);
        return(FALSE);
    }

    if (!(FileHandle = (BPTR) Open(pathname, MODE_OLDFILE)))
    {   DisplayBeep(NULL);
        return(FALSE);
    }
    if (Read(FileHandle, IOBuffer, (LONG) size) != size)
    {   DISCARD Close(FileHandle);
        // FileHandle = NULL;
        return(FALSE);
    }
    DISCARD Close(FileHandle);
    // FileHandle = NULL;

    cursor = 0;
    if (game == BT1)
    {   load_bt1();
    } elif (game == BT2)
    {   load_bt2();
    } else
    {   assert(game == BT3);
        load_bt3();
    }
    refreshgadgets(1);

    return(TRUE);
}

MODULE SWORD getword(void)
{   SWORD value;

    value = ((     256 * IOBuffer[cursor])
           +             IOBuffer[cursor + 1]
            );
    cursor += 2;
    return(value);
}
// Little-endian Intel bullshit.
MODULE SWORD getiword(void)
{   SWORD value;

    value = (            IOBuffer[cursor]
           +      (256 * IOBuffer[cursor + 1])
            );
    cursor += 2;
    return(value);
}
MODULE UBYTE getfatbyte(void)
{   UBYTE value;

    value =              IOBuffer[cursor + 1];
    cursor += 2;
    return(value);
}
MODULE SLONG getlong(void)
{   SLONG value;

    value = ((16777216 * IOBuffer[cursor])
           + (   65536 * IOBuffer[cursor + 1])
           + (     256 * IOBuffer[cursor + 2])
           +             IOBuffer[cursor + 3]
            );
    cursor += 4;
    return(value);
}
MODULE SLONG getilong(void)
{   SLONG value;

    value = (            IOBuffer[cursor]
           + (     256 * IOBuffer[cursor + 1])
           + (   65536 * IOBuffer[cursor + 2])
           + (16777216 * IOBuffer[cursor + 3])
            );
    cursor += 4;
    return(value);
}
MODULE UBYTE getbyte(void)
{   UBYTE value;

    value = IOBuffer[cursor];
    cursor++;
    return(value);
}

MODULE void load_bt1(void)
{   ULONG i;
    UBYTE tempbyte;

    if (platform == PLATFORM_IBM)
    {   strncpy(charname, IOBuffer, 14); //  0-13
        cursor += 3;                 // 14-16
    } else
    {   assert(platform == PLATFORM_AMIGA);
        cursor++;
    }
    tempbyte = getbyte();            // 1 or 17
    filetovar_status(tempbyte);

    if (platform == PLATFORM_IBM)
    {   cursor++; // 18

        tempbyte = getiword();       // 19-20
        filetovar_race(tempbyte);

        tempbyte = getiword();       // 21-22
        filetovar_class(tempbyte);

        curst  = getiword(); // 23-24
        curiq  = getiword(); // 25-26
        curdx  = getiword(); // 27-28
        curcn  = getiword(); // 29-30
        curlk  = getiword(); // 31-32
        maxst  = getiword(); // 33-34
        maxiq  = getiword(); // 35-36
        maxdx  = getiword(); // 37-38
        maxcn  = getiword(); // 39-40
        maxlk  = getiword(); // 41-42
        armour = getiword(); // 43-44
        maxhp  = getiword(); // 45-46
        curhp  = getiword(); // 47-48
        maxsp  = getiword(); // 49-50
        cursp  = getiword(); // 51-52
    } else
    {   assert(platform == PLATFORM_AMIGA);

        tempbyte = getword();        //  2- 3
        filetovar_race(tempbyte);

        tempbyte = getword();        //  4- 5
        filetovar_class(tempbyte);

        curst  = getword(); //  6- 7
        curiq  = getword(); //  8- 9
        curdx  = getword(); // 10-11
        curcn  = getword(); // 12-13
        curlk  = getword(); // 14-15
        maxst  = getword(); // 16-17
        maxiq  = getword(); // 18-19
        maxdx  = getword(); // 20-21
        maxcn  = getword(); // 22-23
        maxlk  = getword(); // 24-25
        armour = getword(); // 26-27
        maxhp  = getword(); // 28-29
        curhp  = getword(); // 30-31
        maxsp  = getword(); // 32-33
        cursp  = getword(); // 34-35
    }

    for (i = 0; i <= 7; i++)
    {   if (getbyte())
        {   equip[i] = TRUE;
        } else
        {   equip[i] = FALSE; // 36, 38, 40, 42, 44, 46, 48, 50 or 53, 55, 57, 59, 61, 63, 65, 67
        }
        item[i] = getbyte();  // 37, 39, 41, 43, 45, 47, 49, 51 or 54, 56, 58, 60, 62, 64, 66, 68
        if (item[i])
        {   quantity[i] = 255;
        } else
        {   quantity[i] = 0;
    }   }
    for (i = 8; i <= 11; i++)
    {   equip[i]    = FALSE;
        item[i]     =
        quantity[i] = 0;
    }

    if (platform == PLATFORM_AMIGA)
    {   xp      = getlong();      //  52- 55
        gp      = getlong();      //  56- 59
        curlev  = getword();      //  60- 61
        maxlev  = getword();      //  62- 63
        sorc    = getword();      //  64- 65
        conj    = getword();      //  66- 67
        magi    = getword();      //  68- 69
        wiza    = getword();      //  70- 71
        cursor  += 9;             //  72- 80
        songs   = getbyte();      //  81
        cursor  += 7;             //  82- 88
        attacks = getword() + 1;  //  89- 90
        // cursor += 5;               91- 95

        make_charname();
    } else
    {   assert(platform == PLATFORM_IBM);
        xp      = getilong();     //  69- 72
        gp      = getilong();     //  73- 76
        curlev  = getiword();     //  77- 78
        maxlev  = getiword();     //  79- 80
        sorc    = getbyte();      //  81
        conj    = getbyte();      //  82
        magi    = getbyte();      //  83
        wiza    = getbyte();      //  84
        cursor  += 8;             //  85- 92
        songs   = getbyte();      //  93
        cursor  += 7;             //  94-100
        attacks = getiword() + 1; // 101-102
        // cursor += 6;              103-108
    }

    arch = damage = image = special = 0;
}

/* Filename conventions:
Amiga BT1: TPW.<name>.C
Amiga BT2: TPW.<name>.C
Amiga BT3: THIEVES.INF
IBM   BT1: <number>.TPW
IBM   BT2: <number>.TW
IBM   BT3: THIEVES.INF */

MODULE void save_bt1(void)
{   ULONG i;

    if (platform == PLATFORM_AMIGA)
    {   for (i = 0; i <= 95; i++)
        {   IOBuffer[i] = 0;
        }

        IOBuffer[1]  = vartofile_status();
        IOBuffer[3]  = vartofile_race();
        IOBuffer[5]  = vartofile_class();

        IOBuffer[7]  = (UBYTE) curst;
        IOBuffer[9]  = (UBYTE) curiq;
        IOBuffer[11] = (UBYTE) curdx;
        IOBuffer[13] = (UBYTE) curcn;
        IOBuffer[15] = (UBYTE) curlk;
        IOBuffer[17] = (UBYTE) maxst;
        IOBuffer[19] = (UBYTE) maxiq;
        IOBuffer[21] = (UBYTE) maxdx;
        IOBuffer[23] = (UBYTE) maxcn;
        IOBuffer[25] = (UBYTE) maxlk;
        IOBuffer[27] = (UBYTE) armour;
        IOBuffer[28] = (UBYTE) (maxhp / 256); // these parentheses are needed
        IOBuffer[29] = (UBYTE) (maxhp % 256);
        IOBuffer[30] = (UBYTE) (curhp / 256);
        IOBuffer[31] = (UBYTE) (curhp % 256);
        IOBuffer[32] = (UBYTE) (maxsp / 256);
        IOBuffer[33] = (UBYTE) (maxsp % 256);
        IOBuffer[34] = (UBYTE) (cursp / 256);
        IOBuffer[35] = (UBYTE) (cursp % 256);

        for (i = 0; i <= 7; i++)
        {   if (equip[i])
            {   IOBuffer[(i * 2) + 36] = 0x80;
            } else
            {   IOBuffer[(i * 2) + 36] = 0;
            }
            IOBuffer[(i * 2) + 37] = item[i];
        }

        IOBuffer[52] = (UBYTE)   (xp / 16777216);
        IOBuffer[53] = (UBYTE)  ((xp % 16777216) / 65536);
        IOBuffer[54] = (UBYTE) (((xp % 16777216) % 65536) / 256);
        IOBuffer[55] = (UBYTE) (((xp % 16777216) % 65536) % 256);
        IOBuffer[56] = (UBYTE)   (gp / 16777216);
        IOBuffer[57] = (UBYTE)  ((gp % 16777216) / 65536);
        IOBuffer[58] = (UBYTE) (((gp % 16777216) % 65536) / 256);
        IOBuffer[59] = (UBYTE) (((gp % 16777216) % 65536) % 256);
        IOBuffer[61] = (UBYTE) curlev;
        IOBuffer[63] = (UBYTE) maxlev;
        IOBuffer[65] = (UBYTE) sorc;
        IOBuffer[67] = (UBYTE) conj;
        IOBuffer[69] = (UBYTE) magi;
        IOBuffer[71] = (UBYTE) wiza;
        IOBuffer[81] = (UBYTE) songs;
        IOBuffer[93] = (UBYTE) (attacks - 1);

        writeout(96);

        make_charname();
        DISCARD SetGadgetAttrs
        (   GadgetsPtr[GID_ST13], MainWindowPtr, NULL,
            STRINGA_TextVal, charname,
        TAG_END);
        RefreshGadgets((struct Gadget *) GadgetsPtr[GID_ST13], MainWindowPtr, NULL);
    } else
    {   assert(platform == PLATFORM_IBM);
        for (i = 0; i <= 108; i++)
        {   IOBuffer[i] = 0;
        }
        strncpy(IOBuffer, charname, 14); // 0..14

        IOBuffer[17] = vartofile_status();
        IOBuffer[19] = vartofile_race();
        IOBuffer[21] = vartofile_class();
        IOBuffer[23] = (UBYTE) curst;
        IOBuffer[25] = (UBYTE) curiq;
        IOBuffer[27] = (UBYTE) curdx;
        IOBuffer[29] = (UBYTE) curcn;
        IOBuffer[31] = (UBYTE) curlk;
        IOBuffer[33] = (UBYTE) maxst;
        IOBuffer[35] = (UBYTE) maxiq;
        IOBuffer[37] = (UBYTE) maxdx;
        IOBuffer[39] = (UBYTE) maxcn;
        IOBuffer[41] = (UBYTE) maxlk;
        IOBuffer[43] = (UBYTE) armour;
        IOBuffer[45] = (UBYTE) (maxhp % 256);
        IOBuffer[46] = (UBYTE) (maxhp / 256); // these parentheses are needed
        IOBuffer[47] = (UBYTE) (curhp % 256);
        IOBuffer[48] = (UBYTE) (curhp / 256);
        IOBuffer[49] = (UBYTE) (maxsp % 256);
        IOBuffer[50] = (UBYTE) (maxsp / 256);
        IOBuffer[51] = (UBYTE) (cursp % 256);
        IOBuffer[52] = (UBYTE) (cursp / 256);

        for (i = 0; i <= 7; i++)
        {   if (equip[i])
            {   IOBuffer[(i * 2) + 53] = 0x80;
            } else
            {   IOBuffer[(i * 2) + 53] = 0;
            }
            IOBuffer[(i * 2) + 54] = item[i];
        }

        IOBuffer[69] = (UBYTE) (((xp % 16777216) % 65536) % 256);
        IOBuffer[70] = (UBYTE) (((xp % 16777216) % 65536) / 256);
        IOBuffer[71] = (UBYTE)  ((xp % 16777216) / 65536);
        IOBuffer[72] = (UBYTE)   (xp / 16777216);
        IOBuffer[73] = (UBYTE) (((gp % 16777216) % 65536) % 256);
        IOBuffer[74] = (UBYTE) (((gp % 16777216) % 65536) / 256);
        IOBuffer[75] = (UBYTE)  ((gp % 16777216) / 65536);
        IOBuffer[76] = (UBYTE)   (gp / 16777216);
        IOBuffer[77] = (UBYTE) curlev;
        IOBuffer[79] = (UBYTE) maxlev;
        IOBuffer[81] = (UBYTE) sorc;
        IOBuffer[82] = (UBYTE) conj;
        IOBuffer[83] = (UBYTE) magi;
        IOBuffer[84] = (UBYTE) wiza;
        IOBuffer[93] = (UBYTE) songs;
        IOBuffer[101] = (UBYTE) (attacks - 1);

        writeout(109);
}   }

MODULE void save_bt3(void)
{   ULONG adder, i;

    for (i = 0; i <= 121; i++)
    {   IOBuffer[i] = 0;
    } // only really need to do 0..119 for IBM-PC

    strncpy(IOBuffer, charname, 14); // 0..14
    // pad byte
    IOBuffer[16] = (UBYTE) maxst;
    IOBuffer[17] = (UBYTE) maxiq;
    IOBuffer[18] = (UBYTE) maxdx;
    IOBuffer[19] = (UBYTE) maxcn;
    IOBuffer[20] = (UBYTE) maxlk;
    // 21 is a pad byte (Amiga only)

    if (platform == PLATFORM_AMIGA)
    {   IOBuffer[22] = (UBYTE)   (xp / 16777216);
        IOBuffer[23] = (UBYTE)  ((xp % 16777216) / 65536);
        IOBuffer[24] = (UBYTE) (((xp % 16777216) % 65536) / 256);
        IOBuffer[25] = (UBYTE) (((xp % 16777216) % 65536) % 256);
        IOBuffer[26] = (UBYTE)   (gp / 16777216);
        IOBuffer[27] = (UBYTE)  ((gp % 16777216) / 65536);
        IOBuffer[28] = (UBYTE) (((gp % 16777216) % 65536) / 256);
        IOBuffer[29] = (UBYTE) (((gp % 16777216) % 65536) % 256);
        IOBuffer[30] = (UBYTE) (curlev / 65536);
        IOBuffer[31] = (UBYTE) (curlev % 65536);
        IOBuffer[32] = (UBYTE) (maxlev / 65536);
        IOBuffer[33] = (UBYTE) (maxlev % 65536);
        IOBuffer[34] = (UBYTE) (curhp / 256);
        IOBuffer[35] = (UBYTE) (curhp % 256);
        IOBuffer[36] = (UBYTE) (maxhp / 256); // these parenthese are needed
        IOBuffer[37] = (UBYTE) (maxhp % 256);
        IOBuffer[38] = (UBYTE) (cursp / 256);
        IOBuffer[39] = (UBYTE) (cursp % 256);
        IOBuffer[40] = (UBYTE) (maxsp / 256);
        IOBuffer[41] = (UBYTE) (maxsp % 256);
        adder = 1;
    } else
    {   assert(platform == PLATFORM_IBM);
        IOBuffer[21] = (UBYTE) (((xp % 16777216) % 65536) % 256);
        IOBuffer[22] = (UBYTE) (((xp % 16777216) % 65536) / 256);
        IOBuffer[23] = (UBYTE)  ((xp % 16777216) / 65536);
        IOBuffer[24] = (UBYTE)   (xp / 16777216);
        IOBuffer[25] = (UBYTE) (((gp % 16777216) % 65536) % 256);
        IOBuffer[26] = (UBYTE) (((gp % 16777216) % 65536) / 256);
        IOBuffer[27] = (UBYTE)  ((gp % 16777216) / 65536);
        IOBuffer[28] = (UBYTE)   (gp / 16777216);
        IOBuffer[29] = (UBYTE) (curlev % 65536);
        IOBuffer[30] = (UBYTE) (curlev / 65536);
        IOBuffer[31] = (UBYTE) (maxlev % 65536);
        IOBuffer[32] = (UBYTE) (maxlev / 65536);
        IOBuffer[33] = (UBYTE) (curhp % 256);
        IOBuffer[34] = (UBYTE) (curhp / 256);
        IOBuffer[35] = (UBYTE) (maxhp % 256); // these parentheses are needed
        IOBuffer[36] = (UBYTE) (maxhp / 256);
        IOBuffer[37] = (UBYTE) (cursp % 256);
        IOBuffer[38] = (UBYTE) (cursp / 256);
        IOBuffer[39] = (UBYTE) (maxsp % 256);
        IOBuffer[40] = (UBYTE) (maxsp / 256);
        adder = 0;
    }

    IOBuffer[41 + adder] = vartofile_class();
    IOBuffer[42 + adder] = vartofile_race();
    IOBuffer[43 + adder] = (UBYTE) sex;
    IOBuffer[44 + adder] = (UBYTE) image;
    IOBuffer[45 + adder] = vartofile_status();
    // 2 pad bytes

    for (i = 0; i <= 11; i++) // 48-83 or 49-84
    {   if (equip[i])
        {   IOBuffer[(i * 3) + 48 + adder] = 1;
        } else
        {   IOBuffer[(i * 3) + 48 + adder] = 0;
        }
        IOBuffer[(i * 3) + 49 + adder] = item[i];
        IOBuffer[(i * 3) + 50 + adder] = quantity[i];
    }

    // spells (84-99 or 85-100)
    for (i = 0; i <= SPELLS; i++)
    {   if (spell[i])
        {   IOBuffer[spells[i].thebyte + adder] |= spells[i].thebit;
    }   }

    IOBuffer[100 + adder] = (UBYTE) songs;       // 100 or 101
    cursor += 5;                                 // 101-104 or 102-105
    IOBuffer[105 + adder] = (UBYTE) attacks - 1; // 105 or 106
    // skip(14 or 15);                           // 107-119 or 108-121

    if (platform == PLATFORM_AMIGA)
    {   adder++; // another pad byte
    }

    IOBuffer[120 + adder] = 0; // the EOF!       // 120 or 122

    if (platform == PLATFORM_IBM)
    {   if (append)
        {   writeout(121);
        } else
        {   writeout(120);
    }   }
    else
    {   assert(platform == PLATFORM_AMIGA);
        if (append)
        {   writeout(123);
        } else
        {   writeout(122);
}   }   }

MODULE void load_bt2(void)
{   UBYTE tempbyte;
    ULONG i;

    if (platform == PLATFORM_AMIGA)
    {   cursor++;             // 0
        tempbyte = getbyte(); // 1
        filetovar_status(tempbyte);
    } else
    {   assert(platform == PLATFORM_IBM);
        strncpy(charname, IOBuffer, 14); // 0..14
        cursor += 17;
        tempbyte = getbyte(); // 17
        filetovar_status(tempbyte);
        cursor++;             // 18
    }

    tempbyte = getbyte(); //  2 or 19
    filetovar_race(tempbyte);

    tempbyte = getbyte(); //  3 or 20
    filetovar_class(tempbyte);

    curst   = getbyte(); //  4 or 21
    curiq   = getbyte(); //  5 or 22
    curdx   = getbyte(); //  6 or 23
    curcn   = getbyte(); //  7 or 24
    curlk   = getbyte(); //  8 or 25
    maxst   = getbyte(); //  9 or 26
    maxiq   = getbyte(); // 10 or 27
    maxdx   = getbyte(); // 11 or 28
    maxcn   = getbyte(); // 12 or 29
    maxlk   = getbyte(); // 13 or 30
    cursor += 2;         // 14-15 or 31-32

    if (platform == PLATFORM_AMIGA)
    {   armour  = getword();  // 16-17
        maxhp   = getword();  // 18-19
        curhp   = getword();  // 20-21
        maxsp   = getword();  // 22-23
        cursp   = getword();  // 24-25
    } else
    {   assert(platform == PLATFORM_IBM);
        armour  = getiword(); // 33-34
        maxhp   = getiword(); // 35-36
        curhp   = getiword(); // 37-38
        maxsp   = getiword(); // 39-40
        cursp   = getiword(); // 41-42
    }
    for (i = 0; i <= 7; i++)
    {   if (getbyte())
        {   equip[i] = TRUE;
        } else equip[i] = FALSE; // 26, 28, 30, 32, 34, 36, 38, 40 or 43, 45, 47, 49, 51, 53, 55, 57
        item[i]     = getbyte(); // 27, 29, 31, 33, 35, 37, 39, 41 or 44, 46, 48, 50, 52, 54, 56, 58
    }
    for (i = 0; i <= 7; i++)
    {   quantity[i] = getbyte(); // 42-49 or 59-66
    }
    for (i = 8; i <= 11; i++)
    {   equip[i]    = FALSE;
        item[i]     =
        quantity[i] = 0;
    }

    if (platform == PLATFORM_AMIGA)
    {   xp      = getlong();  // 50-53
        gp      = getlong();  // 54-57

        make_charname();
    } else
    {   assert(platform == PLATFORM_IBM);
        xp      = getilong(); // 67-70
        gp      = getilong(); // 71-74
    }
    curlev  = getbyte(); // 58 or 75
    maxlev  = getbyte(); // 59 or 76
    sorc    = getbyte(); // 60 or 77
    conj    = getbyte(); // 61 or 78
    magi    = getbyte(); // 62 or 79
    wiza    = getbyte(); // 63 or 80
    arch    = getbyte(); // 64 or 81
    cursor += 4;         // 65-68 or 82-85
    songs   = getbyte(); // 69 or 86
    cursor += 3;         // 70-72 or 87-89
    attacks = getbyte() + 1; // 73 or 90
    cursor += 9;         // 74-82 or 91-99
    image   = getbyte(); // 83 or 100

    tempbyte = getbyte(); // 84 or 101
    switch(tempbyte)
    {
    case 1:
        special =  7; // poison
    acase 2:
        special =  5; // level drain
    acase 3:
        special =  3; // insanity
    acase 4:
        special =  1; // aging
    acase 5:
        special =  8; // possession
    acase 6:
        special =  9; // stoned
    acase 7:
        special =  2; // critical hit
    acase 8:
        special =  4; // item-zot
    acase 9:
        special =  6; // point phaze
    adefault: // eg. 0
        special =  0; // normal
    break;
    }

    damage  = getbyte(); // 85 or 102
}
MODULE void save_bt2(void)
{   ULONG i;

    if (platform == PLATFORM_AMIGA)
    {   for (i = 0; i <= 85; i++)
        {   IOBuffer[i] = 0;
        }

        if (theclass == CLASSGAD_MONSTER || theclass == CLASSGAD_ILLUSION)
        {   IOBuffer[0]  = 1;
        } else IOBuffer[0] = 0;

        IOBuffer[1]  = vartofile_status();
        IOBuffer[2]  = vartofile_race();
        IOBuffer[3]  = vartofile_class();
        IOBuffer[4]  = (UBYTE) curst;
        IOBuffer[5]  = (UBYTE) curiq;
        IOBuffer[6]  = (UBYTE) curdx;
        IOBuffer[7]  = (UBYTE) curcn;
        IOBuffer[8]  = (UBYTE) curlk;
        IOBuffer[9]  = (UBYTE) maxst;
        IOBuffer[10] = (UBYTE) maxiq;
        IOBuffer[11] = (UBYTE) maxdx;
        IOBuffer[12] = (UBYTE) maxcn;
        IOBuffer[13] = (UBYTE) maxlk;
        IOBuffer[17] = (UBYTE) armour;
        IOBuffer[18] = (UBYTE) (maxhp / 256); // these parentheses are needed
        IOBuffer[19] = (UBYTE) (maxhp % 256);
        IOBuffer[20] = (UBYTE) (curhp / 256);
        IOBuffer[21] = (UBYTE) (curhp % 256);
        IOBuffer[22] = (UBYTE) (maxsp / 256);
        IOBuffer[23] = (UBYTE) (maxsp % 256);
        IOBuffer[24] = (UBYTE) (cursp / 256);
        IOBuffer[25] = (UBYTE) (cursp % 256);

        for (i = 0; i <= 7; i++)
        {   if (equip[i])
            {   IOBuffer[(i * 2) + 26] = 0x80;
            } else
            {   IOBuffer[(i * 2) + 26] = 0;
            }
            IOBuffer[(i * 2) + 27] = item[i];
        }
        for (i = 0; i <= 7; i++)
        {   IOBuffer[i + 42]       = quantity[i];
        }

        IOBuffer[50] = (UBYTE)   (xp / 16777216);
        IOBuffer[51] = (UBYTE)  ((xp % 16777216) / 65536);
        IOBuffer[52] = (UBYTE) (((xp % 16777216) % 65536) / 256);
        IOBuffer[53] = (UBYTE) (((xp % 16777216) % 65536) % 256);
        IOBuffer[54] = (UBYTE)   (gp / 16777216);
        IOBuffer[55] = (UBYTE)  ((gp % 16777216) / 65536);
        IOBuffer[56] = (UBYTE) (((gp % 16777216) % 65536) / 256);
        IOBuffer[57] = (UBYTE) (((gp % 16777216) % 65536) % 256);
        IOBuffer[58] = (UBYTE) curlev;
        IOBuffer[59] = (UBYTE) maxlev;
        IOBuffer[60] = (UBYTE) sorc;
        IOBuffer[61] = (UBYTE) conj;
        IOBuffer[62] = (UBYTE) magi;
        IOBuffer[63] = (UBYTE) wiza;
        IOBuffer[64] = (UBYTE) arch;
        IOBuffer[69] = (UBYTE) songs;
        IOBuffer[73] = (UBYTE) attacks - 1;
        IOBuffer[83] = (UBYTE) image;

        switch(special)
        {
        case 1:
            IOBuffer[84] = 4; // aging
        acase 2:
            IOBuffer[84] = 7; // critical hit
        acase 3:
            IOBuffer[84] = 3; // insanity
        acase 4:
            IOBuffer[84] = 8; // item-zot
        acase 5:
            IOBuffer[84] = 2; // level drain
        acase 6:
            IOBuffer[84] = 9; // point phaze
        acase 7:
            IOBuffer[84] = 1; // poison
        acase 8:
            IOBuffer[84] = 5; // possession
        acase 9:
            IOBuffer[84] = 6; // stoned
        adefault: // eg. 0
            IOBuffer[84] = 0; // normal
        break;
        }

        IOBuffer[85] = (UBYTE) damage;

        writeout(86);

        make_charname();
        DISCARD SetGadgetAttrs
        (   GadgetsPtr[GID_ST13], MainWindowPtr, NULL,
            STRINGA_TextVal, charname,
        TAG_END);
        RefreshGadgets((struct Gadget *) GadgetsPtr[GID_ST13], MainWindowPtr, NULL);
    } else
    {   assert(platform == PLATFORM_IBM);
        for (i = 0; i <= 102; i++)
        {   IOBuffer[i] = 0;
        }
        strncpy(IOBuffer, charname, 14); // 0..13

        if (theclass == CLASSGAD_MONSTER || theclass == CLASSGAD_ILLUSION)
        {   IOBuffer[16]  = 1;
        } else IOBuffer[16] = 0;

        IOBuffer[17] = vartofile_status();
        IOBuffer[19] = vartofile_race();
        IOBuffer[20] = vartofile_class();
        IOBuffer[21] = (UBYTE) curst;
        IOBuffer[22] = (UBYTE) curiq;
        IOBuffer[23] = (UBYTE) curdx;
        IOBuffer[24] = (UBYTE) curcn;
        IOBuffer[25] = (UBYTE) curlk;
        IOBuffer[26] = (UBYTE) maxst;
        IOBuffer[27] = (UBYTE) maxiq;
        IOBuffer[28] = (UBYTE) maxdx;
        IOBuffer[29] = (UBYTE) maxcn;
        IOBuffer[30] = (UBYTE) maxlk;
        IOBuffer[33] = (UBYTE) armour;
        IOBuffer[35] = (UBYTE) (maxhp % 256);
        IOBuffer[36] = (UBYTE) (maxhp / 256); // these parentheses are needed
        IOBuffer[37] = (UBYTE) (curhp % 256);
        IOBuffer[38] = (UBYTE) (curhp / 256);
        IOBuffer[39] = (UBYTE) (maxsp % 256);
        IOBuffer[40] = (UBYTE) (maxsp / 256);
        IOBuffer[41] = (UBYTE) (cursp % 256);
        IOBuffer[42] = (UBYTE) (cursp / 256);

        for (i = 0; i <= 7; i++)
        {   if (equip[i])
            {   IOBuffer[(i * 2) + 43] = 0x80;
            } else
            {   IOBuffer[(i * 2) + 43] = 0;
            }
            IOBuffer[(i * 2) + 44] = item[i];
        }
        for (i = 0; i <= 7; i++)
        {   IOBuffer[i + 59]       = quantity[i];
        }

        IOBuffer[67] = (UBYTE)   (xp / 16777216);
        IOBuffer[68] = (UBYTE)  ((xp % 16777216) / 65536);
        IOBuffer[69] = (UBYTE) (((xp % 16777216) % 65536) / 256);
        IOBuffer[70] = (UBYTE) (((xp % 16777216) % 65536) % 256);
        IOBuffer[71] = (UBYTE)   (gp / 16777216);
        IOBuffer[72] = (UBYTE)  ((gp % 16777216) / 65536);
        IOBuffer[73] = (UBYTE) (((gp % 16777216) % 65536) / 256);
        IOBuffer[74] = (UBYTE) (((gp % 16777216) % 65536) % 256);
        IOBuffer[75] = (UBYTE) curlev;
        IOBuffer[76] = (UBYTE) maxlev;
        IOBuffer[77] = (UBYTE) sorc;
        IOBuffer[78] = (UBYTE) conj;
        IOBuffer[79] = (UBYTE) magi;
        IOBuffer[80] = (UBYTE) wiza;
        IOBuffer[81] = (UBYTE) arch;
        IOBuffer[86] = (UBYTE) songs;
        IOBuffer[90] = (UBYTE) attacks - 1;
        IOBuffer[100] = (UBYTE) image;

        switch(special)
        {
        case 1:
            IOBuffer[101] = 4; // aging
        acase 2:
            IOBuffer[101] = 7; // critical hit
        acase 3:
            IOBuffer[101] = 3; // insanity
        acase 4:
            IOBuffer[101] = 8; // item-zot
        acase 5:
            IOBuffer[101] = 2; // level drain
        acase 6:
            IOBuffer[101] = 9; // point phaze
        acase 7:
            IOBuffer[101] = 1; // poison
        acase 8:
            IOBuffer[101] = 5; // possession
        acase 9:
            IOBuffer[101] = 6; // stoned
        adefault: // eg. 0
            IOBuffer[101] = 0; // normal
        break;
        }

        IOBuffer[102] = (UBYTE) damage;

        writeout(103);
}   }
MODULE void load_bt3(void)
{   UBYTE tempbyte;
    ULONG adder, i;

    armour  = 10;
    damage  =
    special =  0;

    strncpy(charname, IOBuffer, 14);
    cursor += 16;

    curst   = getbyte(); // 16
    curiq   = getbyte(); // 17
    curdx   = getbyte(); // 18
    curcn   = getbyte(); // 19
    curlk   = getbyte(); // 20
    maxst   = curst;
    maxiq   = curiq;
    maxdx   = curdx;
    maxcn   = curcn;
    maxlk   = curlk;

    if (platform == PLATFORM_AMIGA)
    {   cursor++;        // 21
        xp      = getlong();  // 22-23
        gp      = getlong();  // 26-29
        curlev  = getword();  // 30-31
        maxlev  = getword();  // 32-33
        curhp   = getword();  // 34-35
        maxhp   = getword();  // 36-37
        cursp   = getword();  // 38-39
        maxsp   = getword();  // 40-41
    } else
    {   assert(platform == PLATFORM_IBM);
        xp      = getilong(); // 21-24
        gp      = getilong(); // 25-28
        curlev  = getiword(); // 29-30
        maxlev  = getiword(); // 31-32
        curhp   = getiword(); // 33-34
        maxhp   = getiword(); // 35-36
        cursp   = getiword(); // 37-38
        maxsp   = getiword(); // 39-40
    }

    tempbyte = getbyte(); //   42 or 41
    filetovar_class(tempbyte);

    tempbyte = getbyte(); //   43 or 42
    filetovar_race(tempbyte);

    sex      = getbyte(); //   44 or 43
    image    = getbyte(); //   45 or 44

    tempbyte = getbyte(); //   46 or 45
    filetovar_status(tempbyte);

    cursor += 2;          //   47-48 or 46-47

    for (i = 0; i <= 11; i++) // 49-84 or 48-83 (36 bytes)
    {   if (getbyte())
        {   equip[i] = TRUE;
        } else
        {   equip[i] = FALSE;
        }
        item[i]     = getbyte();
        quantity[i] = getbyte();
    }

    // spells (85-100 or 84-99)
    if (platform == PLATFORM_AMIGA)
    {   adder = 1;
    } else
    {   assert(platform == PLATFORM_IBM);
        adder = 0;
    }
    for (i = 0; i <= SPELLS; i++)
    {   if (IOBuffer[spells[i].thebyte + adder] & spells[i].thebit)
        {   spell[i] = TRUE;
        } else
        {   spell[i] = FALSE;
    }   }

    songs   = getbyte();     // 101 or 100

    cursor += 5;             // 102-105 or 101-104
    attacks = getbyte() + 1; // 106     or 105
    // skip(15 or 14);       // 107-121 or 106-119
}

MODULE void bt1_to_bt2(void)
{   UBYTE i, j;

    game = BT1;
    readgadgets();

    for (i = 0; i <= 7; i++)
    {   if (item[i])
        {   for (j = 1; j <= ITEMS; j++)
            {   if (item[i] == items[j].bt1)
                {   item[i] = items[j].bt2;
                    if (item[i] == 0)
                    {   quantity[i] = 0;
                    } else
                    {   quantity[i] = 255;
                    }
                    break;
    }   }   }   }
    // condense();

    game = BT2;
    refreshgadgets(1);
}
MODULE void bt1_to_bt3(void)
{   UBYTE i, j;

    game = BT1;
    readgadgets();

    curst = maxst;
    curiq = maxiq;
    curdx = maxdx;
    curcn = maxcn;
    curlk = maxlk;

    for (i = 0; i <= SPELLS; i++)
    {   if
        (   (spells[i].theclass == CLASSGAD_MAGICIAN && spells[i].level <= magi)
         || (spells[i].theclass == CLASSGAD_CONJURER && spells[i].level <= conj)
         || (spells[i].theclass == CLASSGAD_SORCERER && spells[i].level <= sorc)
         || (spells[i].theclass == CLASSGAD_WIZARD   && spells[i].level <= wiza)
        )
        {   spell[i] = TRUE;
        } else
        {   spell[i] = FALSE;
    }   }
    magi =
    conj =
    sorc =
    wiza = 0; // arch will be 0 anyway

    for (i = 0; i <= 7; i++)
    {   if (item[i])
        {   for (j = 1; j <= ITEMS; j++)
            {   if (item[i] == items[j].bt1)
                {   item[i] = items[j].bt3;
                    if (item[i] == 0)
                    {   quantity[i] = 0;
                    } else
                    {   quantity[i] = 255;
                    }
                    break;
    }   }   }   }
    // condense();

    game = BT3;
    refreshgadgets(1);
}
MODULE void bt2_to_bt1(void)
{   UBYTE i, j;

    game = BT2;
    readgadgets();

    if (theclass == CLASSGAD_ARCHMAGE)
    {   theclass = CLASSGAD_WIZARD;
    } elif (theclass == CLASSGAD_MONSTER || theclass == CLASSGAD_ILLUSION)
    {   theclass = CLASSGAD_WARRIOR;
        race = RACEGAD_HUMAN;
        maxst = maxiq = maxdx = maxcn = maxlk =
        curst = curiq = curdx = curcn = curlk = 10;
        maxlev = curlev = 1;
        xp = songs = 0;
        // We assume no need to change spell levels, etc.
    }

    arch = damage = image = special = 0;
    attacks = 1;

    for (i = 0; i <= 7; i++)
    {   if (item[i])
        {   for (j = 1; j <= ITEMS; j++)
            {   if (item[i] == items[j].bt2)
                {   item[i] = items[j].bt1;
                    if (item[i] == 0)
                    {   quantity[i] = 0;
                    } else
                    {   quantity[i] = 255;
                    }
                    break;
    }   }   }   }
    // condense();

    game = BT1;
    refreshgadgets(1);
}
MODULE void bt2_to_bt3(void)
{   UBYTE i, j;

    game = BT2;
    readgadgets();

    curst = maxst;
    curiq = maxiq;
    curdx = maxdx;
    curcn = maxcn;
    curlk = maxlk;

    for (i = 0; i <= SPELLS; i++)
    {   if
        (   (spells[i].theclass == CLASSGAD_MAGICIAN && spells[i].level <= magi)
         || (spells[i].theclass == CLASSGAD_CONJURER && spells[i].level <= conj)
         || (spells[i].theclass == CLASSGAD_SORCERER && spells[i].level <= sorc)
         || (spells[i].theclass == CLASSGAD_WIZARD   && spells[i].level <= wiza)
         || (spells[i].theclass == CLASSGAD_ARCHMAGE && spells[i].level <= arch)
        )
        {   spell[i] = TRUE;
        } else
        {   spell[i] = FALSE;
    }   }
    magi =
    conj =
    sorc =
    wiza =
    arch = 0;

    if (theclass != CLASSGAD_MONSTER && theclass != CLASSGAD_ILLUSION)
    {   for (i = 0; i <= 7; i++)
        {   if (item[i])
            {   for (j = 1; j <= ITEMS; j++)
                {   if (item[i] == items[j].bt2)
                    {   item[i] = items[j].bt3;
                        if (item[i] == 0)
                        {   quantity[i] = 0;
                        }
                        break;
    }   }   }   }   }
    else
    {   for (i = 0; i <= 7; i++)
        {   item[i]     =
            equip[i]    = FALSE;
            quantity[i] = 0;
    }   }
    // condense();

    game = BT3;
    refreshgadgets(1);
}
MODULE void bt3_to_bt1(void)
{   UBYTE i, j;

    game = BT3;
    readgadgets();

    if (theclass == CLASSGAD_ARCHMAGE || theclass == CLASSGAD_CHRONOMANCER)
    {   theclass = CLASSGAD_WIZARD;
    } elif (theclass == CLASSGAD_GEOMANCER)
    {   theclass = CLASSGAD_WARRIOR;
    } elif (theclass == CLASSGAD_MONSTER || theclass == CLASSGAD_ILLUSION)
    {   theclass = CLASSGAD_WARRIOR;
        race = RACEGAD_HUMAN;
        maxst = maxiq = maxdx = maxcn = maxlk =
        curst = curiq = curdx = curcn = curlk = 10;
        maxlev = curlev = 1;
        xp = songs = 0;
        // We assume no need to change anything else
    }

    arch = damage = image = special = sex = 0;
    attacks = 1;

    conj =
    magi =
    sorc =
    wiza = 0;
    for (i = 0; i <= SPELLS; i++)
    {   if (spell[i])
        {   if     (spells[i].theclass == CLASSGAD_CONJURER && spells[i].level > conj)
            {   conj = spells[i].level;
            } elif (spells[i].theclass == CLASSGAD_MAGICIAN && spells[i].level > magi)
            {   magi = spells[i].level;
            } elif (spells[i].theclass == CLASSGAD_SORCERER && spells[i].level > sorc)
            {   sorc = spells[i].level;
            } elif (spells[i].theclass == CLASSGAD_WIZARD   && spells[i].level > wiza)
            {   wiza = spells[i].level;
    }   }   }

    for (i = 0; i <= 7; i++)
    {   if (item[i])
        {   for (j = 1; j <= ITEMS; j++)
            {   if (item[i] == items[j].bt3)
                {   item[i] = items[j].bt1;
                    if (item[i] == 0)
                    {   quantity[i] = 0;
                    } else
                    {   quantity[i] = 255;
                    }
                    break;
    }   }   }   }
    // condense();

    for (i = 8; i <= 11; i++)
    {   item[i] = quantity[i] = equip[i] = 0;
    }

    game = BT1;
    refreshgadgets(1);
}
MODULE void bt3_to_bt2(void)
{   UBYTE i, j;

    game = BT3;
    readgadgets();

    if (theclass == CLASSGAD_CHRONOMANCER)
    {   theclass = CLASSGAD_ARCHMAGE;
    } elif (theclass == CLASSGAD_GEOMANCER)
    {   theclass = CLASSGAD_WARRIOR;
    }

    sex = 0;

    conj =
    magi =
    sorc =
    wiza =
    arch = 0;
    for (i = 0; i <= SPELLS; i++)
    {   if (spell[i])
        {   if     (spells[i].theclass == CLASSGAD_CONJURER && spells[i].level > conj)
            {   conj = spells[i].level;
            } elif (spells[i].theclass == CLASSGAD_MAGICIAN && spells[i].level > magi)
            {   magi = spells[i].level;
            } elif (spells[i].theclass == CLASSGAD_SORCERER && spells[i].level > sorc)
            {   sorc = spells[i].level;
            } elif (spells[i].theclass == CLASSGAD_WIZARD   && spells[i].level > wiza)
            {   wiza = spells[i].level;
            } elif (spells[i].theclass == CLASSGAD_ARCHMAGE && spells[i].level > arch)
            {   arch = spells[i].level;
    }   }   }

    for (i = 0; i <= 7; i++)
    {   if (item[i])
        {   for (j = 1; j <= ITEMS; j++)
            {   if (item[i] == items[j].bt3)
                {   item[i] = items[j].bt2;
                    if (item[i] == 0)
                    {   quantity[i] = 0;
                    }
                    break;
    }   }   }   }
    // condense();

    for (i = 8; i <= 11; i++)
    {   item[i] = quantity[i] = equip[i] = 0;
    }

    game = BT2;
    refreshgadgets(1);
}

/* MODULE void condense(void)
{   ULONG i, j;

    for (i = 0; i <= 6; i++)
    {   if (item[i] == 0)
        {   for (j = 7; j >= i + 1; j--)
            {   item[j - 1] = item[j];
                quantity[j - 1] = quantity[j];
}   }   }   } */

MODULE LONG getsize(void)
{   BPTR                  FileHandle /* = NULL */ ;
    LONG                  size;
    struct FileInfoBlock* FIBPtr;

    if (!(FileHandle = (BPTR) Lock(pathname, ACCESS_READ)))
    {   return(0);
    }
    if (!(FIBPtr = AllocDosObject(DOS_FIB, NULL)))
    {   UnLock(FileHandle);
        // FileHandle = NULL;
        return(0);
    }
    if (!(Examine(FileHandle, FIBPtr)))
    {   FreeDosObject(DOS_FIB, FIBPtr);
        // FIBPtr = NULL;
        UnLock(FileHandle);
        // FileHandle = NULL;
        return(0);
    }
    size = (ULONG) FIBPtr->fib_Size;
    FreeDosObject(DOS_FIB, FIBPtr);
    // FIBPtr = NULL;
    UnLock(FileHandle);
    // FileHandle = NULL;
    return(size);
}

int main(int argc, char** argv)
{   SLONG                   args[1 + 1] = {0L, 0L};
    UWORD                   code;
    ULONG                   i, result, signals, MainSignal;
    struct ListBrowserNode* NodePtr;

    /* Start of program.

    version embedding into executable */

    if (0) /* that is, never */
    {   DISCARD Printf("$VER: BTCE 2.1a (9.6.2005)"); /* always d.m.y format */
    }

    NewList(&ItemsList); // before the first possible point of failure
    NewList(&SpellsList);

    /* These libraries are used:
    asl, dos, exec, gadtools, intuition, resource, version

    We are happy with any versions of asl, listbrowser. */

    // Check for OS3.5+...
    if
    (     IntuitionBase->LibNode.lib_Version < OS_31
     ||         SysBase->LibNode.lib_Version < OS_31
     ||    GadToolsBase->lib_Version         < OS_21
     ||        IconBase->lib_Version         < OS_35
     ||     VersionBase->lib_Version         < OS_35
     || !(ButtonBase      = OpenLibrary("gadgets/button.gadget",      OS_35 ))
     || !(CheckBoxBase    = OpenLibrary("gadgets/checkbox.gadget",    OS_35 ))
     || !(ChooserBase     = OpenLibrary("gadgets/chooser.gadget",     OS_35 ))
     || !(LabelBase       = OpenLibrary("images/label.image",         OS_35 ))
     || !(LayoutBase      = OpenLibrary("gadgets/layout.gadget",      OS_35 ))
     || !(ListBrowserBase = OpenLibrary("gadgets/listbrowser.gadget", OS_ANY))
     || !(SpaceBase       = OpenLibrary("gadgets/space.gadget",       OS_ANY))
     || !(StringBase      = OpenLibrary("gadgets/string.gadget",      OS_35 ))
     || !(WindowBase      = OpenLibrary("window.class",               OS_35 ))
     || !(IntegerBase     = OpenLibrary("gadgets/integer.gadget",     OS_35 ))
     || !(ResourceBase    = OpenLibrary("resource.library",           OS_ANY))
    )
    {   DISCARD Printf("Need AmigaOS 3.5+!\n");
        cleanexit(EXIT_FAILURE);
    }

    for (i = 0; i <= ITEMS; i++)
    {   NodePtr = (struct ListBrowserNode *) AllocListBrowserNode
        (   1, LBNCA_Text, items[i].name, TAG_END);
        // check NodePtr is non-zero
        AddTail(&ItemsList, (struct Node *) NodePtr);
    }
    for (i = 0; i <= SPELLS; i++)
    {   NodePtr = (struct ListBrowserNode *) AllocListBrowserNode
        (   1,
            LBNCA_Text,    spells[i].longname,
            LBNA_CheckBox, TRUE,
        TAG_END);
        // check NodePtr is non-zero
        AddTail(&SpellsList, (struct Node *) NodePtr);
    }

    if (argc) // started from CLI
    {   if (!(ArgsPtr = ReadArgs
        (   "FILE/F", //  0
            (LONG *) args,
            NULL
        )))
        {   DISCARD Printf
            (   "Usage: %s [[FILE] <filename>]\n",
                argv[0]
            );
            cleanexit(EXIT_FAILURE);
        }
        if (args[0])
        {   strcpy(pathname, (STRPTR) args[0]);
    }   }
    // else started from WB

    lockscreen();
    if (GetVPModeID(&(ScreenPtr->ViewPort)) == INVALID_ID)
    {   Printf("Invalid default public screen mode ID!\n");
        cleanexit(EXIT_FAILURE);
    }
    if (!(VisualInfoPtr = (struct VisualInfo *) GetVisualInfo(ScreenPtr, TAG_DONE)))
    {   Printf("Can't get GadTools visual info!\n");
        cleanexit(EXIT_FAILURE);
    }
    if (!(MenuPtr = CreateMenus(NewMenu, TAG_DONE)))
    {   Printf("Can't create menus!\n");
        cleanexit(EXIT_FAILURE);
    }
    if (!(LayoutMenus(MenuPtr, VisualInfoPtr, GTMN_NewLookMenus, TRUE, TAG_DONE)))
    {   Printf("Can't lay out menus!\n");
        cleanexit(EXIT_FAILURE);
    }
    if (!(ResourcePtr = RL_OpenResource(RCTResource, ScreenPtr, NULL)))
    {   rq("RL_OpenResource() failed!");
    }
    if (!(WinObject = RL_NewObject
    (   ResourcePtr,
        WIN_1_ID,
        WINDOW_MenuStrip, MenuPtr,
        WA_PubScreen,     ScreenPtr,
    TAG_END)))
    {   rq("RL_NewObject() failed!");
    }
    if (!(GadgetsPtr = (struct Gadget **) RL_GetObjectArray(ResourcePtr, WinObject, GROUP_2_ID)))
    {   rq("RL_GetObjectArray() failed!");
    }
    DoMethod(WinObject, WM_OPEN);
    DISCARD GetAttr(WINDOW_Window,  WinObject, (ULONG *) &MainWindowPtr);
    unlockscreen();

    // spells gadget
    DISCARD SetGadgetAttrs(GadgetsPtr[GID_LB2], MainWindowPtr, NULL, LISTBROWSER_Labels, ~0,          TAG_END);
    assert(SpellsList.lh_Head->ln_Succ); // the list is non-empty
    DISCARD SetGadgetAttrs(GadgetsPtr[GID_LB2], MainWindowPtr, NULL, LISTBROWSER_Labels, &SpellsList, TAG_END);
    RefreshGadgets((struct Gadget *) GadgetsPtr[GID_LB2], MainWindowPtr, NULL);

    DISCARD ActivateLayoutGadget(GadgetsPtr[GID_LY1], MainWindowPtr, NULL, (Object) GadgetsPtr[GID_IN1]);

    /* get the signalbit for the windowmessages */
    GetAttr(WINDOW_SigMask, WinObject, &MainSignal);

    if (pathname[0])
    {   load();
        refreshgadgets(1);
    } else
    {   project_new();
    }

    while (1)
    {   /* wait for a message or CTRL-C for user break */
        signals = Wait(MainSignal | SIGBREAKF_CTRL_C);

        /* handle the window-messages */
        if (signals & MainSignal)
        {   /* get next message */
            /* result: high word is the class, in low word is the event-describe */

            while((result = RA_HandleInput(WinObject, &code)) != WMHI_LASTMSG)
            {   switch(result & WMHI_CLASSMASK)
                {
                case WMHI_CLOSEWINDOW:
                    cleanexit(EXIT_SUCCESS);
                acase WMHI_GADGETUP:
                    /* Note: we use RL_GADGETMASK instead of WMHI_GADGETMASK
                     * because ReActor ors the gadget ID and the group ID to
                     * the final gadget ID. If you want to know in which group
                     * the gadget was use (result & RL_GROUPMASK).
                     */
                    loop(result & RL_GADGETMASK);
                acase WMHI_MENUPICK:
                    handlemenus(code);
                acase WMHI_RAWKEY:
                    /* raw key event, lower byte are the keycode, qualifiers only in hook-routine */
                    switch(code & 0xFF)
                    {
                    /* too easy to trigge accidentally!
                    case SCAN_ESCAPE:
                        if (!(qual & IEQUALIFIER_REPEAT))
                        {   cleanexit(EXIT_SUCCESS);
                        }
                    */
                    case SCAN_HELP:
                        about();
                    adefault:
                    break;
                    }
                acase WMHI_VANILLAKEY:
                    /* ASCII-code in the lower byte (see RawKey) */
                    // we never seem to receive these messages
                adefault:
                break;
        }   }   }

        /* handle CTRL-C signal */
        if (signals & SIGBREAKF_CTRL_C)
        {   cleanexit(EXIT_SUCCESS);
    }   }

    // control should never reach this point
    assert(0);
}

MODULE void cleanexit(SBYTE rc)
{   // ASL requesters and About windows are assumed to be already closed.

    unlockscreen(); // in case the screen is locked

    if (MainWindowPtr)
    {   clearkybd();

        /* Disposing of the window object will also close the window if it is
         * already opened, and it will dispose of the layout object attached to it.
         */
        RL_DisposeObject(ResourcePtr, WinObject);
        WinObject = NULL;
        MainWindowPtr = NULL;
    }

    if (ResourcePtr)
    {   RL_CloseResource(ResourcePtr);
        ResourcePtr = NULL;
    }
    if (ItemsList.lh_Head->ln_Succ) // if list is non-empty
    {   FreeListBrowserList(&ItemsList);
    }
    if (SpellsList.lh_Head->ln_Succ) // if list is non-empty
    {   FreeListBrowserList(&SpellsList);
    }
    if (MenuPtr)
    {   FreeMenus(MenuPtr);
        MenuPtr = NULL;
    }
    if (VisualInfoPtr)
    {   FreeVisualInfo(VisualInfoPtr);
        VisualInfoPtr = NULL;
    }
    if (ArgsPtr)
    {   FreeArgs(ArgsPtr);
        ArgsPtr = NULL;
    }

    if      (ButtonBase) CloseLibrary(ButtonBase);
    if    (CheckBoxBase) CloseLibrary(CheckBoxBase);
    if     (ChooserBase) CloseLibrary(ChooserBase);
    if     (IntegerBase) CloseLibrary(IntegerBase);
    if       (LabelBase) CloseLibrary(LabelBase);
    if      (LayoutBase) CloseLibrary(LayoutBase);
    if (ListBrowserBase) CloseLibrary(ListBrowserBase);
    if       (SpaceBase) CloseLibrary(SpaceBase);
    if      (StringBase) CloseLibrary(StringBase);
    if      (WindowBase) CloseLibrary(WindowBase);
    if    (ResourceBase) CloseLibrary(ResourceBase);

    if (IntuitionBase)
    {   DISCARD OpenWorkBench();
    }

    exit(rc); // End of program.
}

MODULE void lockscreen(void)
{   if (!(ScreenPtr = LockPubScreen(NULL)))
    {   Printf("Can't lock default public screen!\n");
        cleanexit(EXIT_FAILURE);
}   }
MODULE void unlockscreen(void)
{   if (ScreenPtr)
    {   UnlockPubScreen(NULL, ScreenPtr);
        ScreenPtr = NULL;
}   }

MODULE void clearkybd(void)
{   struct IntuiMessage* MsgPtr;

    while ((MsgPtr = (struct IntuiMessage *) GT_GetIMsg(MainWindowPtr->UserPort)))
        GT_ReplyIMsg(MsgPtr);
}

MODULE void refreshgadgets(ULONG mode)
{   ULONG        i, j;
    FLAG         found = FALSE;
    struct Node* NodePtr;
    struct List* BrowserListPtr;

    /* Modes are:

    0: inventory only
    1: everything
    2: spells only */

if (mode == 1)
{   // general

    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_ST13], MainWindowPtr, NULL,
        STRINGA_TextVal, charname,
        GA_Disabled, (platform == PLATFORM_AMIGA && game != BT3),
    TAG_END);
    RefreshGadgets((struct Gadget *) GadgetsPtr[GID_ST13], MainWindowPtr, NULL);

    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_CH1], MainWindowPtr, NULL,
        CHOOSER_Selected, (WORD) game,
    TAG_END);
    RefreshGadgets((struct Gadget *) GadgetsPtr[GID_CH1], MainWindowPtr, NULL);

    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_CH6], MainWindowPtr, NULL,
        CHOOSER_Selected, (WORD) platform,
    TAG_END);
    RefreshGadgets((struct Gadget *) GadgetsPtr[GID_CH6], MainWindowPtr, NULL);

    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_IN1], MainWindowPtr, NULL,
        INTEGER_Number, gp,
    TAG_END);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_IN2], MainWindowPtr, NULL,
        INTEGER_Number, curhp,
    TAG_END);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_IN3], MainWindowPtr, NULL,
        INTEGER_Number, maxhp,
    TAG_END);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_IN4], MainWindowPtr, NULL,
        INTEGER_Number, cursp,
    TAG_END);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_IN5], MainWindowPtr, NULL,
        INTEGER_Number, maxsp,
    TAG_END);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_CH5], MainWindowPtr, NULL,
        CHOOSER_Selected, (WORD) statusgad,
    TAG_END);
    RefreshGadgets((struct Gadget *) GadgetsPtr[GID_CH5], MainWindowPtr, NULL);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_IN21], MainWindowPtr, NULL,
        INTEGER_Number, armour,
        GA_Disabled,    game == BT3,
    TAG_END);

    // attributes
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_IN6], MainWindowPtr, NULL,
        INTEGER_Number, curst,
        GA_Disabled,    game == BT3,
    TAG_END);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_IN7], MainWindowPtr, NULL,
        INTEGER_Number, curiq,
        GA_Disabled,    game == BT3,
    TAG_END);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_IN8], MainWindowPtr, NULL,
        INTEGER_Number, curdx,
        GA_Disabled,    game == BT3,
    TAG_END);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_IN9], MainWindowPtr, NULL,
        INTEGER_Number, curcn,
        GA_Disabled,    game == BT3,
    TAG_END);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_IN10], MainWindowPtr, NULL,
        INTEGER_Number, curlk,
        GA_Disabled,    game == BT3,
    TAG_END);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_IN11], MainWindowPtr, NULL,
        INTEGER_Number, maxst,
    TAG_END);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_IN12], MainWindowPtr, NULL,
        INTEGER_Number, maxiq,
    TAG_END);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_IN13], MainWindowPtr, NULL,
        INTEGER_Number, maxdx,
    TAG_END);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_IN14], MainWindowPtr, NULL,
        INTEGER_Number, maxcn,
    TAG_END);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_IN15], MainWindowPtr, NULL,
        INTEGER_Number, maxlk,
    TAG_END);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_IN16], MainWindowPtr, NULL,
        INTEGER_Number, magi,
        GA_Disabled,    game == BT3,
    TAG_END);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_IN17], MainWindowPtr, NULL,
        INTEGER_Number, conj,
        GA_Disabled,    game == BT3,
    TAG_END);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_IN18], MainWindowPtr, NULL,
        INTEGER_Number, sorc,
        GA_Disabled,    game == BT3,
    TAG_END);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_IN19], MainWindowPtr, NULL,
        INTEGER_Number, wiza,
        GA_Disabled,    game == BT3,
    TAG_END);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_IN20], MainWindowPtr, NULL,
        INTEGER_Number, arch,
        GA_Disabled,    game != BT2,
    TAG_END);

    // illusions/monsters
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_IN22], MainWindowPtr, NULL,
        INTEGER_Number,   attacks,
    TAG_END);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_IN23], MainWindowPtr, NULL,
        INTEGER_Number,   damage,
        GA_Disabled,      game != BT2,
    TAG_END);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_BU1], MainWindowPtr, NULL,
        BUTTON_Integer,   damage * 4,
    TAG_END);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_IN24], MainWindowPtr, NULL,
        INTEGER_Number,   image,
        GA_Disabled,      game == BT1,
    TAG_END);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_CH3], MainWindowPtr, NULL,
        CHOOSER_Selected, (WORD) special,
        GA_Disabled,      game != BT2,
    TAG_END);
    RefreshGadgets((struct Gadget *) GadgetsPtr[GID_CH3], MainWindowPtr, NULL);

    // players
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_CH4], MainWindowPtr, NULL,
        CHOOSER_Selected, (WORD) race,
    TAG_END);
    RefreshGadgets((struct Gadget *) GadgetsPtr[GID_CH4], MainWindowPtr, NULL);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_IN25], MainWindowPtr, NULL,
        INTEGER_Number, curlev,
    TAG_END);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_IN26], MainWindowPtr, NULL,
        INTEGER_Number, maxlev,
    TAG_END);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_IN27], MainWindowPtr, NULL,
        INTEGER_Number, songs,
    TAG_END);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_IN28], MainWindowPtr, NULL,
        INTEGER_Number, xp,
    TAG_END);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_CH5], MainWindowPtr, NULL,
        CHOOSER_Selected, statusgad,
    TAG_END);
    RefreshGadgets((struct Gadget *) GadgetsPtr[GID_CH5], MainWindowPtr, NULL);
    DISCARD SetGadgetAttrs
    (   GadgetsPtr[GID_CH2], MainWindowPtr, NULL,
        CHOOSER_Selected, (WORD) sex,
        GA_Disabled,      game != BT3,
    TAG_END);
    RefreshGadgets((struct Gadget *) GadgetsPtr[GID_CH2], MainWindowPtr, NULL);

    // invert selection
    DISCARD SetGadgetAttrs(GadgetsPtr[GID_BU14], MainWindowPtr, NULL, GA_Disabled,            game != BT3,     TAG_END);
    // buttons autorefresh

    // class
    DISCARD GetAttr(LISTBROWSER_Labels, GadgetsPtr[GID_LB3], (ULONG *) &BrowserListPtr);
    DISCARD SetGadgetAttrs(GadgetsPtr[GID_LB3], MainWindowPtr, NULL, LISTBROWSER_Labels, ~0, TAG_END);
    assert(BrowserListPtr->lh_Head->ln_Succ); // the list is non-empty
    i = 0;
    // Walk the list
    for
    (   NodePtr = BrowserListPtr->lh_Head;
        NodePtr->ln_Succ;
        NodePtr = NodePtr->ln_Succ
    )
    {   if
        (   i == CLASSGAD_ARCHMAGE
         || i == CLASSGAD_ILLUSION
         || i == CLASSGAD_MONSTER
        )
        {   if (game == BT1)
            {   DISCARD SetListBrowserNodeAttrs(NodePtr, LBNA_Flags, LBFLG_READONLY, TAG_END);
            } else
            {   DISCARD SetListBrowserNodeAttrs(NodePtr, LBNA_Flags, NULL,           TAG_END);
        }   }
        elif
        (   i == CLASSGAD_CHRONOMANCER
         || i == CLASSGAD_GEOMANCER
        )
        {   if (game != BT3)
            {   DISCARD SetListBrowserNodeAttrs(NodePtr, LBNA_Flags, LBFLG_READONLY, TAG_END);
            } else
            {   DISCARD SetListBrowserNodeAttrs(NodePtr, LBNA_Flags, NULL,           TAG_END);
        }   }
        i++;
    }
    DISCARD SetGadgetAttrs(GadgetsPtr[GID_LB3], MainWindowPtr, NULL, LISTBROWSER_Labels,      BrowserListPtr,  TAG_END);
    DISCARD SetGadgetAttrs(GadgetsPtr[GID_LB3], MainWindowPtr, NULL, LISTBROWSER_Selected,    (WORD) theclass, TAG_END);
    DISCARD SetGadgetAttrs(GadgetsPtr[GID_LB3], MainWindowPtr, NULL, LISTBROWSER_MakeVisible, theclass,  TAG_END);
    RefreshGadgets((struct Gadget *) GadgetsPtr[GID_LB3], MainWindowPtr, NULL);

    if (game == BT3)
    {    OnMenu(MainWindowPtr, FULLMENUNUM(MN_SETTINGS, IN_APPEND, NOSUB));
    } else
    {   OffMenu(MainWindowPtr, FULLMENUNUM(MN_SETTINGS, IN_APPEND, NOSUB));
}   }

if (mode >= 1)
{   // spells
    DISCARD GetAttr(LISTBROWSER_Labels, GadgetsPtr[GID_LB2], (ULONG *) &BrowserListPtr);
    DISCARD SetGadgetAttrs(GadgetsPtr[GID_LB2], MainWindowPtr, NULL, GA_Disabled,             game != BT3,     TAG_END);
    DISCARD SetGadgetAttrs(GadgetsPtr[GID_LB2], MainWindowPtr, NULL, LISTBROWSER_Labels,      ~0,              TAG_END);
    assert(BrowserListPtr->lh_Head->ln_Succ); // the list is non-empty
    i = 0;
    // Walk the list
    for
    (   NodePtr = BrowserListPtr->lh_Head;
        NodePtr->ln_Succ;
        NodePtr = NodePtr->ln_Succ
    )
    {   DISCARD SetListBrowserNodeAttrs(NodePtr, LBNA_Checked, (BOOL) spell[i], TAG_END);
        i++;
    }
    DISCARD SetGadgetAttrs(GadgetsPtr[GID_LB2], MainWindowPtr, NULL, LISTBROWSER_Labels,      BrowserListPtr,  TAG_END);
    DISCARD SetGadgetAttrs(GadgetsPtr[GID_LB2], MainWindowPtr, NULL, LISTBROWSER_Selected,    (WORD) theclass, TAG_END);
    DISCARD SetGadgetAttrs(GadgetsPtr[GID_LB2], MainWindowPtr, NULL, LISTBROWSER_MakeVisible, theclass,  TAG_END);
    RefreshGadgets((struct Gadget *) GadgetsPtr[GID_LB2], MainWindowPtr, NULL);
}

if (mode <= 1)
{   // items
    for (i = 0; i <= 11; i++)
    {   DISCARD SetGadgetAttrs
        (   GadgetsPtr[gad_cb[i]], MainWindowPtr, NULL,
            GA_Selected, equip[i],
            GA_Disabled, (game != BT3 && i >= 8),
        TAG_END);
        RefreshGadgets((struct Gadget *) GadgetsPtr[gad_cb[i]], MainWindowPtr, NULL);

        if (item[i])
        {   found = FALSE;

            for (j = 1; j <= ITEMS; j++)
            {   if
                (   (game == BT1 && items[j].bt1 == item[i])
                 || (game == BT2 && items[j].bt2 == item[i])
                 || (game == BT3 && items[j].bt3 == item[i])
                )
                {   DISCARD SetGadgetAttrs
                    (   GadgetsPtr[gad_st[i]], MainWindowPtr, NULL,
                        STRINGA_TextVal, items[j].name,
                    TAG_END);
                    found = TRUE;
                    break;
            }   }

            if (!found)
            {   // we should not reach this point normally
                DISCARD SetGadgetAttrs
                (   GadgetsPtr[gad_st[i]], MainWindowPtr, NULL,
                    STRINGA_TextVal, "?",
                TAG_END);
        }   }
        else
        {   DISCARD SetGadgetAttrs
            (   GadgetsPtr[gad_st[i]], MainWindowPtr, NULL,
                STRINGA_TextVal, "-",
            TAG_END);
        }
        DISCARD SetGadgetAttrs
        (   GadgetsPtr[gad_in[i]], MainWindowPtr, NULL,
            INTEGER_Number, quantity[i],
            GA_Disabled,    (game == BT1 || (game == BT2 && i >= 8)),
        TAG_END);
        DISCARD SetGadgetAttrs
        (   GadgetsPtr[gad_bu[i]], MainWindowPtr, NULL,
            GA_Disabled,    (game != BT3 && i >= 8),
        TAG_END);
}   }
}

MODULE void rq(STRPTR message)
{   Printf("%s\n", message);
    cleanexit(EXIT_FAILURE);
}

MODULE void writeout(SLONG length)
{   BPTR FileHandle;

    if (!(FileHandle = (BPTR) Open(pathname, MODE_NEWFILE)))
        rq("Can't open file for writing!");
    if (Write(FileHandle, IOBuffer, length) != length)
    {   DISCARD Close(FileHandle);
        rq("Can't write to file!");
    }
    DISCARD Close(FileHandle);
}

MODULE void open(void)
{   if (asl(FALSE))
    {   load();
        refreshgadgets(TRUE);
}   }
MODULE void save(void)
{   readgadgets();

    if (asl(TRUE))
    {   if (game == BT1)
        {   save_bt1();
        } elif (game == BT2)
        {   save_bt2();
        } else
        {   assert(game == BT3);
            if (warn())
            {   save_bt3();
}   }   }   }

MODULE FLAG asl(FLAG save)
{   struct FileRequester* ASLRqPtr;
    FLAG                  success;
    ULONG                 flags;
    LONG                  length;
    TEXT                  hail[20 + 1],
                          tempstring[14 + 1];

    if (save)
    {   flags = FILF_PATGAD | FILF_SAVE;
        if (game == BT1)
        {   strcpy(hail, "Save BT1 Character");
        } elif (game == BT2)
        {   strcpy(hail, "Save BT2 Character");
        } else
        {   assert(game == BT3);
            strcpy(hail, "Save BT3 Character");
        }
    } else
    {   flags = FILF_PATGAD;
        strcpy(hail, "Load BT1-3 Character");
    }

    /* path is the directory that the ASL requester will start in. */

    if (!(ASLRqPtr = AllocAslRequestTags(ASL_FileRequest, ASL_Pattern, "#?", ASL_Window, MainWindowPtr, TAG_DONE)))
        rq("Can't create ASL request!");
    if
    (   AslRequestTags
        (   ASLRqPtr,
            ASLFR_InitialDrawer, path,
            ASLFR_InitialFile,   filename,
            ASLFR_TitleText,     hail,
            ASLFR_Flags1,        flags,
        TAG_DONE)
     && *(ASLRqPtr->rf_File) != 0
    )
    {   strcpy(path,     ASLRqPtr->rf_Dir);
        strcpy(pathname, path);
        strcpy(filename, ASLRqPtr->rf_File);

        if (save)
        {   if (game != BT3 && platform == PLATFORM_AMIGA)
            {   if (strnicmp(filename, "TPW.", 4))
                {   strcpy(tempstring, filename);
                    strcpy(filename, "TPW.");
                    strcat(filename, tempstring);
                    strcat(filename, ".C");
            }   }
            elif (game == BT1)
            {   length = strlen(filename);
                if (length < 5 || strnicmp(&filename[length - 4], ".TPW", 4))
                {   strcat(filename, ".TPW");
            }   }
            elif (game == BT2)
            {   length = strlen(filename);
                if (length < 4 || strnicmp(&filename[length - 3], ".TW",  3))
                {   strcat(filename, ".TW");
        }   }   }

        if (!AddPart(pathname, filename, 768))
        {   FreeAslRequest(ASLRqPtr);
            rq("AddPart() failed!");
        }

        success = TRUE;
    } else
    {   // either the user chose Cancel, or clicked OK with an empty filename
        filename[0] =
        pathname[0] = 0;
        success = FALSE;
    }
    assert((int) ASLRqPtr);
    FreeAslRequest(ASLRqPtr);

    return(success);
}

MODULE void handlemenus(UWORD code)
{   struct MenuItem* ItemPtr;

    if (code != MENUNULL)
    {   ItemPtr = ItemAddress(MenuPtr, code);
        switch (MENUNUM(code))
        {
        case MN_PROJECT:
            switch (ITEMNUM(code))
            {
            case IN_NEW:
                project_new();
            acase IN_OPEN:
                open();
            acase IN_SAVEAS:
                save();
            acase IN_QUIT:
                cleanexit(EXIT_SUCCESS);
            adefault:
            break;
            }
        acase MN_SETTINGS:
            switch (ITEMNUM(code))
            {
            case IN_APPEND:
                if (ItemPtr->Flags & CHECKED)
                {   append = TRUE;
                } else
                {   append = FALSE;
                }
            adefault:
            break;
            }
        acase MN_HELP:
            switch (ITEMNUM(code))
            {
            case IN_ABOUT:
                about();
            adefault:
            break;
            }
        adefault:
        break;
}   }   }

MODULE void loop(ULONG gid)
{   ULONG i;

    switch(gid)
    {
    case GID_CH1:
        if (game == BT1)
        {   DISCARD GetAttr(CHOOSER_Selected, GadgetsPtr[GID_CH1],  (ULONG *) &game);
            if (game == BT2)
            {   bt1_to_bt2();
            } elif (game == BT3)
            {   bt1_to_bt3();
        }   }
        elif (game == BT2)
        {   DISCARD GetAttr(CHOOSER_Selected, GadgetsPtr[GID_CH1],  (ULONG *) &game);
            if (game == BT1)
            {   bt2_to_bt1();
            } elif (game == BT3)
            {   bt2_to_bt3();
        }   }
        else
        {   assert(game == BT3);
            DISCARD GetAttr(CHOOSER_Selected, GadgetsPtr[GID_CH1],  (ULONG *) &game);
            if (game == BT1)
            {   bt3_to_bt1();
            } elif (game == BT2)
            {   bt3_to_bt2();
        }   }
    acase GID_LB3:
        DISCARD GetAttr(LISTBROWSER_Selected, GadgetsPtr[GID_LB3],  (ULONG *) &theclass);

        if (game == BT3 && (theclass == CLASSGAD_MONSTER || theclass == CLASSGAD_ILLUSION))
        {   for (i = 0; i <= 11; i++)
            {   item[i]     =
                equip[i]    = FALSE;
                quantity[i] = 0;
        }   }
    acase GID_CH6:
        DISCARD GetAttr(CHOOSER_Selected, GadgetsPtr[GID_CH6],  (ULONG *) &platform);
        refreshgadgets(TRUE); // to enable/disable "character name" as appropriate
    acase GID_IN23:
        DISCARD GetAttr(INTEGER_Number,       GadgetsPtr[GID_IN23], (ULONG *) &damage);
        DISCARD SetGadgetAttrs
        (   GadgetsPtr[GID_BU1], MainWindowPtr, NULL,
            BUTTON_Integer, damage * 4,
        TAG_END);
        RefreshGadgets((struct Gadget *) GadgetsPtr[GID_BU1], MainWindowPtr, NULL);
    acase GID_BU2:
        itemwindow(0);
    acase GID_BU3:
        itemwindow(1);
    acase GID_BU4:
        itemwindow(2);
    acase GID_BU5:
        itemwindow(3);
    acase GID_BU6:
        itemwindow(4);
    acase GID_BU7:
        itemwindow(5);
    acase GID_BU8:
        itemwindow(6);
    acase GID_BU9:
        itemwindow(7);
    acase GID_BU10:
        itemwindow(8);
    acase GID_BU11:
        itemwindow(9);
    acase GID_BU12:
        itemwindow(10);
    acase GID_BU13:
        itemwindow(11);
    acase GID_BU14:
        readgadgets();
        for (i = 0; i <= SPELLS; i++)
        {   if (spell[i])
            {   spell[i] = FALSE;
            } else
            {   spell[i] = TRUE;
            }
        }
        refreshgadgets(2);
    adefault:
    break;
}   }

MODULE void itemwindow(ULONG whichitem)
{   FLAG            done;
    UWORD           code;
    ULONG           event,
                    i,
                    ItemSignal,
                    ordinal,
                    result;
    LONG            white;
    struct Window*  ItemWindowPtr;
    struct Node*    NodePtr;
    struct Gadget** ItemGadgetsPtr;
    Object*         ItemWinObject;

    lockscreen();
    if (!(ItemWinObject = RL_NewObject
    (   ResourcePtr,
        WIN_5_ID,
        WA_PubScreen, ScreenPtr,
    TAG_END)))
    {   rq("RL_NewObject() failed!");
    }
    if (!(ItemGadgetsPtr = (struct Gadget **) RL_GetObjectArray(ResourcePtr, ItemWinObject, GROUP_6_ID)))
    {   rq("RL_GetObjectArray() failed!");
    }
    DoMethod(ItemWinObject, WM_OPEN);
    DISCARD GetAttr(WINDOW_Window, ItemWinObject, (ULONG *) &ItemWindowPtr);
    white = FindColor
    (   ScreenPtr->ViewPort.ColorMap,
        0xFFFFFFFF,
        0xFFFFFFFF,
        0xFFFFFFFF,
        -1
    );
    unlockscreen();

    DISCARD SetGadgetAttrs(ItemGadgetsPtr[GID_LB1], ItemWindowPtr, NULL, LISTBROWSER_Labels, ~0,         TAG_END);
    DISCARD SetGadgetAttrs(ItemGadgetsPtr[GID_LB1], ItemWindowPtr, NULL, LISTBROWSER_Labels, &ItemsList, TAG_END);
    for (i = 0; i <= ITEMS; i++)
    {   if
        (   (game == BT1 && item[whichitem] == items[i].bt1)
         || (game == BT2 && item[whichitem] == items[i].bt2)
         || (game == BT3 && item[whichitem] == items[i].bt3)
        )
        {   DISCARD SetGadgetAttrs(ItemGadgetsPtr[GID_LB1], ItemWindowPtr, NULL, LISTBROWSER_Selected, i, TAG_END);
            ordinal = i;
            break;
    }   }
    RefreshGadgets((struct Gadget *) ItemGadgetsPtr[GID_LB1], ItemWindowPtr, NULL);

    DISCARD SetGadgetAttrs(ItemGadgetsPtr[GID_LB1], ItemWindowPtr, NULL, LISTBROWSER_Labels, ~0, TAG_END);
    assert(ItemsList.lh_Head->ln_Succ); // the list is non-empty
    i = 0;
    // Walk the list

    for
    (   NodePtr = ItemsList.lh_Head;
        NodePtr->ln_Succ;
        NodePtr = NodePtr->ln_Succ
    )
    {   if (game == BT1)
        {   if (i == 0 || items[i].bt1)
            {   DISCARD SetListBrowserNodeAttrs(NodePtr, LBNA_Flags, NULL, TAG_END);
            } else
            {   DISCARD SetListBrowserNodeAttrs(NodePtr, LBNA_Flags, LBFLG_READONLY | LBFLG_CUSTOMPENS, TAG_END);
                DISCARD SetListBrowserNodeAttrs(NodePtr, LBNA_Column, 0, LBNCA_FGPen, white, TAG_END);
        }   }
        elif (game == BT2)
        {   if (i == 0 || items[i].bt2)
            {   DISCARD SetListBrowserNodeAttrs(NodePtr, LBNA_Flags, NULL, TAG_END);
            } else
            {   DISCARD SetListBrowserNodeAttrs(NodePtr, LBNA_Flags, LBFLG_READONLY | LBFLG_CUSTOMPENS, TAG_END);
                DISCARD SetListBrowserNodeAttrs(NodePtr, LBNA_Column, 0, LBNCA_FGPen, white, TAG_END);
        }   }
        else
        {   assert(game == BT3);
            if (i == 0 || items[i].bt3)
            {   DISCARD SetListBrowserNodeAttrs(NodePtr, LBNA_Flags, NULL, TAG_END);
            } else
            {   DISCARD SetListBrowserNodeAttrs(NodePtr, LBNA_Flags, LBFLG_READONLY | LBFLG_CUSTOMPENS, TAG_END);
                DISCARD SetListBrowserNodeAttrs(NodePtr, LBNA_Column, 0, LBNCA_FGPen, white, TAG_END);
        }   }

        i++;
    }
    DISCARD SetGadgetAttrs(ItemGadgetsPtr[GID_LB1], ItemWindowPtr, NULL, LISTBROWSER_Labels,      &ItemsList, TAG_END);
    DISCARD SetGadgetAttrs(ItemGadgetsPtr[GID_LB1], ItemWindowPtr, NULL, LISTBROWSER_MakeVisible, ordinal,  TAG_END);
    RefreshGadgets((struct Gadget *) ItemGadgetsPtr[GID_LB1], ItemWindowPtr, NULL);

    // Obtain the window wait signal mask.
    DISCARD GetAttr(WINDOW_SigMask, ItemWinObject, (ULONG *) &ItemSignal);

    done = FALSE;
    while (!done)
    {   if ((Wait(ItemSignal | SIGBREAKF_CTRL_C)) & SIGBREAKF_CTRL_C)
        {   DoMethod(ItemWinObject, WM_CLOSE);
            RL_DisposeObject(ResourcePtr, ItemWinObject);
            cleanexit(EXIT_SUCCESS);
        }
        while ((result = DoMethod(ItemWinObject, WM_HANDLEINPUT, &code)) != WMHI_LASTMSG)
        {   /* There seems to be a bug in StormC affecting the code
            generation:

            Some switch statements can cause a "bxx.b L0" to
            be generated by the compiler. No label L0 is defined, thus
            we get "Panic: Assembler failed at n: Undefined Label". */
            event = result & WMHI_CLASSMASK;
            if (event == WMHI_CLOSEWINDOW)
            {   done = TRUE;
            } elif (event == WMHI_GADGETUP)
            {   /* Note: we use RL_GADGETMASK instead of WMHI_GADGETMASK
                 * because ReActor ors the gadget ID and the group ID to
                 * the final gadget ID. If you want to know in which group
                 * the gadget was use (result & RL_GROUPMASK).
                 */
                item_loop(result & RL_GADGETMASK, code, whichitem);
            } elif (event == WMHI_RAWKEY)
            {   /* raw key event, lower byte are the keycode, qualifiers only in hook-routine */
                switch(code & 0xFF)
                {
                case SCAN_ESCAPE:
                    done = TRUE;
                acase SCAN_UP:
                    DISCARD SetGadgetAttrs(ItemGadgetsPtr[GID_LB1], ItemWindowPtr, NULL, LISTBROWSER_Position, LBP_LINEUP,   TAG_END);
                acase SCAN_DOWN:
                    DISCARD SetGadgetAttrs(ItemGadgetsPtr[GID_LB1], ItemWindowPtr, NULL, LISTBROWSER_Position, LBP_LINEDOWN, TAG_END);
                adefault:
                break;
    }   }   }   }

    DoMethod(ItemWinObject, WM_CLOSE);
    RL_DisposeObject(ResourcePtr, ItemWinObject);
}

MODULE void about(void)
{   FLAG           done;
    UWORD          code;
    ULONG          AboutSignal,
                   event,
                   result;
    Object*        AboutWinObject = NULL;
    struct Window* AboutWindowPtr = NULL;

    lockscreen();
    if (!(AboutWinObject = RL_NewObject
    (   ResourcePtr,
        WIN_3_ID,
        WA_PubScreen, ScreenPtr,
    TAG_END)))
    {   rq("RL_NewObject() failed!");
    }
    DoMethod(AboutWinObject, WM_OPEN);
    DISCARD GetAttr(WINDOW_Window,  AboutWinObject, (ULONG *) &AboutWindowPtr);
    unlockscreen();

    /* This should really have a ReActor-style loop rather than
       a ReAction-style one, although it seems to work. */

    // Obtain the window wait signal mask.
    DISCARD GetAttr(WINDOW_SigMask, AboutWinObject, (ULONG *) &AboutSignal);

    done = FALSE;
    while (!done)
    {   if ((Wait(AboutSignal | SIGBREAKF_CTRL_C)) & SIGBREAKF_CTRL_C)
        {   DoMethod(AboutWinObject, WM_CLOSE);
            RL_DisposeObject(ResourcePtr, AboutWinObject);
            cleanexit(EXIT_SUCCESS);
        }
        while ((result = DoMethod(AboutWinObject, WM_HANDLEINPUT, &code)) != WMHI_LASTMSG)
        {   /* There seems to be a bug in StormC affecting the code
            generation:

            Some switch statements can cause a "bxx.b L0" to
            be generated by the compiler. No label L0 is defined, thus
            we get "Panic: Assembler failed at n: Undefined Label". */
            event = result & WMHI_CLASSMASK;
            if (event == WMHI_CLOSEWINDOW)
            {   done = TRUE;
            } elif (event == WMHI_RAWKEY)
            {   /* raw key event, lower byte are the keycode, qualifiers only in hook-routine */
                switch(code & 0xFF)
                {
                case SCAN_ESCAPE:
                    done = TRUE;
                adefault:
                break;
    }   }   }   }

    DoMethod(AboutWinObject, WM_CLOSE);
    RL_DisposeObject(ResourcePtr, AboutWinObject);
}

MODULE void filetovar_race(UBYTE thebyte)
{   switch(thebyte)
    {
    case 1:
        race = 1; // elf
    acase 2:
        race = 0; // dwarf
    acase 3:
        race = 5; // hobbit
    acase 4:
        race = 3; // half-elf
    acase 5:
        race = 4; // half-orc
    acase 6:
        race = 2; // gnome
    adefault: // eg. 0
        race = 6; // human
    break;
}   }
MODULE void filetovar_class(UBYTE thebyte)
{   switch(thebyte)
    {
    case 1:
        if (game == BT3)
        {   theclass = CLASSGAD_WIZARD;
        } else theclass = CLASSGAD_PALADIN;
    acase 2:
        if (game == BT3)
        {   theclass = CLASSGAD_SORCERER;
        } else theclass = CLASSGAD_ROGUE;
    acase 3:
        if (game == BT3)
        {   theclass = CLASSGAD_CONJURER;
        } else theclass = CLASSGAD_BARD;
    acase 4:
        if (game == BT3)
        {   theclass = CLASSGAD_MAGICIAN;
        } else theclass = CLASSGAD_HUNTER;
    acase 5:
        if (game == BT3)
        {   theclass = CLASSGAD_ROGUE;
        } else theclass = CLASSGAD_MONK;
    acase 6:
        if (game == BT3)
        {   theclass = CLASSGAD_BARD;
        } else theclass = CLASSGAD_CONJURER;
    acase 7:
        if (game == BT3)
        {   theclass = CLASSGAD_PALADIN;
        } else theclass = CLASSGAD_MAGICIAN;
    acase 8:
        if (game == BT3)
        {   theclass = CLASSGAD_HUNTER;
        } else theclass = CLASSGAD_SORCERER;
    acase 9:
        if (game == BT3)
        {   theclass = CLASSGAD_MONK;
        } else theclass = CLASSGAD_WIZARD;
    acase 10:
        theclass = CLASSGAD_ARCHMAGE; // same for all games
    acase 11:
        if (game == BT3)
        {   theclass = CLASSGAD_CHRONOMANCER;
        } else // game should be BT2
        {   theclass = CLASSGAD_MONSTER;
        }
    acase 12:
        if (game == BT3)
        {   theclass = CLASSGAD_GEOMANCER;
        } else // game should be BT2
        {   theclass = CLASSGAD_ILLUSION;
        }
    acase 13:
        // game should be BT3
        theclass = CLASSGAD_MONSTER;
    acase 14:
        // game should be BT3
        theclass = CLASSGAD_ILLUSION;
    adefault: // eg. 0
        theclass = CLASSGAD_WARRIOR; // same for all games
    break;
}   }

MODULE UBYTE vartofile_race(void)
{   UBYTE thebyte;

    switch(race)
    {
    case 0:
        thebyte = 2; // dwarf
    acase 1:
        thebyte = 1; // elf
    acase 2:
        thebyte = 6; // gnome
    acase 3:
        thebyte = 4; // half-elf
    acase 4:
        thebyte = 5; // half-orc
    acase 5:
        thebyte = 3; // hobbit
    acase 6:
        thebyte = 0; // human
    adefault:
        assert(0);
    break;
    }

    return(thebyte);
}
MODULE UBYTE vartofile_class(void)
{   UBYTE thebyte;

    switch(theclass)
    {
    case CLASSGAD_ARCHMAGE:
        thebyte = 10; // same for all games
    acase CLASSGAD_BARD:
        if (game == BT3) thebyte =  6; else thebyte =  3;
    acase CLASSGAD_CHRONOMANCER:
        assert(game == BT3);
        thebyte = 11;
    acase CLASSGAD_CONJURER:
        if (game == BT3) thebyte =  3; else thebyte =  6;
    acase CLASSGAD_GEOMANCER:
        assert(game == BT3);
        thebyte = 12;
    acase CLASSGAD_HUNTER:
        if (game == BT3) thebyte =  8; else thebyte =  4;
    acase CLASSGAD_ILLUSION:
        if (game == BT3) thebyte = 14; else thebyte = 12;
    acase CLASSGAD_MAGICIAN:
        if (game == BT3) thebyte =  4; else thebyte =  7;
    acase CLASSGAD_MONK:
        if (game == BT3) thebyte =  9; else thebyte =  5;
    acase CLASSGAD_MONSTER:
        if (game == BT3) thebyte = 13; else thebyte = 11;
    acase CLASSGAD_PALADIN:
        if (game == BT3) thebyte =  7; else thebyte =  1;
    acase CLASSGAD_ROGUE:
        if (game == BT3) thebyte =  5; else thebyte =  2;
    acase CLASSGAD_SORCERER:
        if (game == BT3) thebyte =  2; else thebyte =  8;
    acase CLASSGAD_WARRIOR:
        thebyte =  0; // same for all games
    acase CLASSGAD_WIZARD:
        if (game == BT3) thebyte =  1; else thebyte =  9;
    adefault:
        assert(0);
    break;
    }

    return(thebyte);
}
MODULE UBYTE vartofile_status(void)
{   UBYTE thebyte;

    switch(statusgad)
    {
    case 0:
        thebyte = 0;    // OK
    acase 1:
        thebyte = 0x02; // dead
    acase 2:
        thebyte = 0x80; // insane
    acase 3:
        thebyte = 0x04; // old
    acase 4:
        thebyte = 0x20; // paralyzed
    acase 5:
        thebyte = 0x08; // poisoned
    acase 6:
        thebyte = 0x40; // possessed
    acase 7:
        thebyte = 0x10; // stoned
    adefault:
        assert(0);
    break;
    }

    return(thebyte);
}
MODULE void filetovar_status(UBYTE thebyte)
{   switch(thebyte)
    {
    case 0x80:
        statusgad = 2; // insane
    acase 0x40:
        statusgad = 6; // possessed
    acase 0x20:
        statusgad = 4; // paralyzed
    acase 0x10:
        statusgad = 7; // stoned
    acase 0x08:
        statusgad = 5; // poisoned
    acase 0x04:
        statusgad = 3; // old
    acase 0x02:
        statusgad = 1; // dead
    adefault:
        // eg. 0, which is perfect health
        statusgad = 0; // OK
    break;
}   }

MODULE void item_loop(ULONG gid, UWORD code, ULONG whichitem)
{   switch(gid)
    {
    case GID_LB1:
        if (code == 0)
        {   quantity[whichitem] = 0;
            equip[whichitem] = FALSE;
        } elif (item[whichitem] == 0)
        {   quantity[whichitem] = 255;
        }
        if (game == BT1)
        {   item[whichitem] = items[code].bt1;
        } elif (game == BT2)
        {   item[whichitem] = items[code].bt2;
        } else
        {   assert(game == BT3);
            item[whichitem] = items[code].bt3;
        }
    adefault:
    break;
    }

    refreshgadgets(0);
}

MODULE void project_new(void)
{   ULONG i;

    // game and platform are unaffected.

    charname[0] = 0;

    statusgad =
    sorc      =
    conj      =
    magi      =
    wiza      =
    arch      =
    songs     =
    image     =
    special   =
    damage    =
    maxsp     =
    cursp     =
    sex       =
    xp        = 0;

    attacks   =
    curlev    =
    maxlev    = 1;

    curst     =
    curiq     =
    curdx     =
    curcn     =
    curlk     =
    maxst     =
    maxiq     =
    maxdx     =
    maxcn     =
    maxlk     =
    armour    =
    maxhp     =
    curhp     = 10;

    gp        = 100;

    for (i = 0; i <= 11; i++)
    {   equip[i]    = FALSE;
        item[i]     = 0;
        quantity[i] = 0;
    }
    for (i = 0; i <= SPELLS; i++)
    {   spell[i]    = FALSE;
    }

    theclass  = CLASSGAD_WARRIOR;
    race      = RACEGAD_HUMAN;

    refreshgadgets(1);
}

#ifdef __STORM__
void wbmain(struct WBStartup* WBMsg)
{   main(0, (char **) WBMsg);
}
#endif

MODULE void make_charname(void)
{   LONG  length;
    ULONG i;

    // make character name from filename
    // For Amiga BT1/BT2.

    if (strnicmp(filename, "TPW.", 4))
    {   // if a strange filename, just use the entire filename as the charname

        strncpy(charname, filename, 14);
    } else
    {   length = strlen(filename);
        for (i = 4; i < length; i++)
        {   if (filename[i] == '.') // normally there is a C after this .
            {   charname[i - 4] = 0;
                return;
            } else
            {   charname[i - 4] = filename[i];
}   }   }   }

MODULE LONG warn(void)
{   return EasyRequest(MainWindowPtr, &EasyStruct, NULL, pathname);

    /* 1 will be returned for "Confirm".
       0 will be returned for "Cancel". */
}
